"""Folder module."""

from typing import Iterable

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.Container import Container
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Feature import Feature
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class Folder(Container):
    """A KML `<Document>` tag constructor.

    A `<Folder>` is a container for `<Feature>` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#folder

    Parameters
    ----------
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Folder` is enabled (checked) in the "Places" panel when
        it is initially loaded. In order for any child `Feature`s to be visible, the
        `<visibility>` tag of all its ancestors (`Document` or `Folder` instances) must
        also be set `True`.
    is_open : bool | None, default = None
        Specifies whether the `Folder` appears closed or open when first loaded into
        the "Places" panel. `False` or `None` is collapsed (the default), `True` is
        expanded.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.

        A given KML `Feature` can contain a combination of these types of custom data.
    features : Feature | Iterable[Feature] | None, default = None
        The `Feature`'s contained by this `Folder`.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Folder"

    def __init__(
        self,
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
        features: Feature | Iterable[Feature] | None = None,
    ):
        """Folder instance constructor."""
        super().__init__(
            name=name,
            visibility=visibility,
            is_open=is_open,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
            features=features,
        )



"""StyleSelector module."""

from abc import ABC

from pyLiveKML.objects.Object import Object


class StyleSelector(Object, ABC):
    """A KML `<StyleSelector>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file. It is the base
    type for `Style` and `StyleMap`. An object derived from `StyleSelector` is uniquely
    identified by its id and its url.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#styleselector.

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    def __init__(self) -> None:
        """StyleSelector instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)



"""TimeStamp module."""

from datetime import datetime

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef, _DateTimeParse
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class TimeStamp(TimePrimitive):
    """A KML `<TimeStamp>` tag constructor.

    Represents a single moment in time. This is a simple element and contains no
    children. Its value is a `datetime`, specified in XML time. The precision of the
    `TimeStamp` is dictated by the value of `when`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#style
    * http://www.w3.org/TR/xmlschema-2/#isoformats

    Parameters
    ----------
    when : datetime | str

    Attributes
    ----------
    when : datetime

    """

    _kml_tag = "TimeStamp"
    _kml_fields = TimePrimitive._kml_fields + (
        _FieldDef("when", parser=_DateTimeParse),
    )

    def __init__(
        self,
        when: datetime | str,
    ):
        """TimeStamp instance constructor."""
        TimePrimitive.__init__(self)
        self.when = when



"""Object module."""

from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Any, Generic, Iterable, Iterator, Type, TypeVar
from uuid import uuid4, UUID

from dateutil.parser import parse as dtparse
from lxml import etree  # type: ignore

from pyLiveKML.types.GeoColor import GeoColor
from pyLiveKML.utils import with_ns


class _KMLDump(ABC):
    """Abstract base for various "Dump" implementations.

    "Dump" classes are used to convert a Python object into something suitable for
    publishing as a KML tag value.

    """

    @classmethod
    @abstractmethod
    def dump(cls, value: Any) -> Any:
        raise NotImplementedError


class _NoDump(_KMLDump):
    """Dump nothing, i.e. an empty string."""

    @classmethod
    def dump(cls, value: Any) -> Any:
        """Dump and format a KML object field."""
        return ""


class _DumpDirect(_KMLDump):
    """Dump to string."""

    @classmethod
    def dump(cls, value: Any) -> Any:
        """Dump and format a KML object field."""
        if value is None:
            return None
        if isinstance(value, Enum):
            return str(value.value)
        if isinstance(value, bool):
            return str(int(value))
        if isinstance(value, datetime):
            return value.isoformat()
        return str(value)


class _KMLParser(ABC):
    """Abstract base for various "Parser" implementations.

    "Parser" classes are used to constrain or otherwise manipulate a parameter passed to
    a `_BaseObject`.

    """

    @classmethod
    @abstractmethod
    def parse(cls, value: Any) -> Any:
        raise NotImplementedError


class _NoParse(_KMLParser):
    """A value that will not be changed."""

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        return value


class _Angle90(_KMLParser):
    """A value in the range -90 to +90.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        value = float(value)
        return 90 if value > 90 else -90 if value < -90 else value


class _AnglePos90(_KMLParser):
    """A value in the range 0 to +90.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        value = float(value)
        return 90 if value > 90 else 0 if value < 0 else value


class _Angle180(_KMLParser):
    """A value in the range -180 to +180.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        value = float(value)
        while value > 180:
            value = value - 360
        while value < -180:
            value = value + 360
        return value


class _AnglePos180(_KMLParser):
    """A value in the range 0 to +180.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        value = float(value)
        return 180 if value > 180 else 0 if value < 0 else value


class _Angle360(_KMLParser):
    """A value in the range -360 to +360.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        value = float(value)
        return (
            value % 360 if value > 360 else -(-value % 360) if value < -360 else value
        )


class _ColorParse(_KMLParser):
    """Ensures that the output is a `GeoColor`, or `None`.

    Primarily intended to convert an `int` to a `GeoColor`.

    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        if isinstance(value, int):
            return GeoColor(value)
        return value


class _DateTimeParse(_KMLParser):
    """Ensures that the output is a `datetime`, or `None`."""

    @classmethod
    def parse(cls, value: Any) -> Any:
        """Transform the argument."""
        if isinstance(value, str):
            return dtparse(value)
        return value


class _FieldDef:
    """Describes how a field of a KML object is to be published.

    KML object class definitions (may) specify a tuple of `_FieldDef` instances as the
    `_kml_fields` class variable.

    Parameters
    ----------
    name : str
        The name of the field, from the perspective of the Python object.
    tag : str | None, default = None
        The text that will be assigned to the KML tag for the field when it is published.
        May include a prefixed and colon-separated namespace, e.g. "gx:option" is valid.
    parser : Type[_KMLParser], default = NoParse
        The parser class that will be used to transform any value assigned to the field.
        This allows e.g. floats to be constrained to an appropriate range.
    dumper : Type[_KMLDump], default = DumpDirect
        The dumper class that will be used to convert and publish the field's value to
        KML.

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        name: str,
        tag: str | None = None,
        parser: Type[_KMLParser] = _NoParse,
        dumper: Type[_KMLDump] = _DumpDirect,
    ):
        """_FieldDef instance constructor."""
        self.name = name
        self.typename = tag if tag is not None else name
        self.parser = parser
        self.dumper = dumper


class _ChildDef:
    """Describes how a child of a KML object is to be published.

    KML object class definitions (may) specify a tuple of `_ChildDef` instances as the
    `_kml_children` class variable.

    Parameters
    ----------
    name : str
        The name of the field, from the perspective of the Python object.

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        name: str,
    ) -> None:
        self.name = name


class _DependentDef:
    """Describes how a dependent object of a KML object is to be published.

    KML object class definitions (may) specify a tuple of `_DependentDef` instances as
    the `_kml_dependents` class variable.

    Parameters
    ----------
    name : str
        The name of the field, from the perspective of the Python object.

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        name: str,
    ):
        self.name = name


class ObjectState(Enum):
    """Enumeration of possible states that objects derived from KML :class:`~pyLiveKML.KMLObjects.Object` may hold.

    The `ObjectState` enumeration is specific to the :mod:`pyLiveKML` package, i.e. it is
    *not* part of the KML specification. It is used exclusively for synchronization.

    """

    IDLE = 0
    CREATING = 1
    CREATED = 2
    CHANGING = 3
    DELETE_CREATED = 4
    DELETE_CHANGED = 5


class _BaseObject(ABC):
    """`_BaseObject` is a private class from which almost every KML tag constructor ultimately derives.

    This includes the `Object` class, which is what KML objects are **meant** to derive
    from. The primary difference between `_BaseObject` and `Object` is that `Object` sets
    the `_suppress_id` class variable `True`, while `_BaseObject` sets it `False`. What
    this means in practice is that KML tags that are constructed from `_BaseObject` do
    not include an `id` attribute, so cannot be targeted by e.g. `<Change>` or `<Delete>`
    tags.

    Note that while `_BaseObject` **suppresses** it's `id` when publishing to KML, it still
    **has** an `id` attribute.

    Apart from this, `_BaseObject` and `Object` are conceptually interchangeable.

    """

    _kml_tag: str = ""
    _kml_fields: tuple[_FieldDef, ...] = tuple()
    _kml_children: tuple[_ChildDef, ...] = tuple()
    _kml_dependents: tuple[_DependentDef, ...] = tuple()
    _suppress_id: bool = True

    def __init__(self) -> None:
        """_BaseObject instance constructor."""
        super().__init__()
        self._id: Any = uuid4()
        self._container: _BaseObject | None = None
        self._state: ObjectState = ObjectState.IDLE

    def __setattr__(self, name: str, value: Any) -> None:
        """_BaseObject __setattr__ override method.

        Kicks in when setting an attribute with the same name as an entry in
        `_kml_fields`.

        If a match is made, executes the field's parser to constrain the
        parameter.

        If a matched field changes its value, flags this by calling `field_changed()`.

        """
        match = next(filter(lambda x: x.name == name, self._kml_fields), None)
        changed = False
        if match is not None:
            value = match.parser.parse(value)
            if value != getattr(self, name, None):
                changed = True
        super().__setattr__(name, value)
        if changed:
            self.field_changed()

    def __eq__(self, value: object) -> bool:
        """_BaseObject __eq__ override method.

        When comparing two _BaseObject instances of the same type, considers them equal
        if all of the attributes listed in `_kml_fields` are of equal value.

        """
        return isinstance(value, self.__class__) and all(
            map(
                lambda x: getattr(self, x.name) == getattr(value, x.name),
                self._kml_fields,
            )
        )

    def __str__(self) -> str:
        """Return the `_BaseObject`'s `_kml_tag` by default."""
        return self._kml_tag

    def __repr__(self) -> str:
        """Return the `_BaseObject`'s string representation by default."""
        return str(self)

    @property
    def id(self) -> str:
        """The unique identifier of this `_BaseObject`."""
        return str(self._id)

    @property
    def state(self) -> ObjectState:
        """The current GEP synchronization state of this `_BaseObject`."""
        return self._state

    @property
    def active(self) -> bool:
        """Whether the instance has been selected for display.

        Parameters
        ----------
        value : bool
            If `True`, activates the `_BaseObject`; if `False`, deactivates it.

        Returns
        -------
        bool
            `True` if the `_BaseObject` has been created in the UI and is not scheduled
            for deletion, otherwise `False`.

        """
        return (
            ObjectState.CREATING,
            ObjectState.CREATED,
            ObjectState.CHANGING,
        ).__contains__(self._state)

    @active.setter
    def active(self, value: bool) -> None:
        self.activate(value)

    @property
    def fields(self) -> Iterator[tuple[str, Any]]:
        """A generator over the `_BaseObject`'s field names and their values.

        Returns
        -------
        Iterator[tuple[str, Any]]
            In each tuple, the first element is the field name, and the second element is
            the field value.

        """
        for f in self._kml_fields:
            f_obj = getattr(self, f.name, None)
            if f_obj is not None:
                yield ((f.name, f_obj))

    @property
    def children(self) -> Iterator["ObjectChild"]:
        """A generator over the children of the instance.

        In this context, children are child objects that the parent *does not* rely upon.

        Returns
        -------
        Iterator[ObjectChild]
            The children, as `ObjectChild` instances.

        """
        for c in self._kml_children:
            c_obj = getattr(self, c.name, None)
            if c_obj is None:
                continue
            elif isinstance(c_obj, _ListObject):
                if not c_obj._yield_if_empty and len(c_obj) <= 0:
                    continue
                elif c_obj._yield_self:
                    yield ObjectChild(self, c_obj)
                else:
                    yield from (ObjectChild(self, cc) for cc in c_obj)
            elif isinstance(c_obj, Iterable):
                yield from (ObjectChild(self, cc) for cc in c_obj)
            else:
                yield ObjectChild(self, c_obj)

    @property
    def dependents(self) -> Iterator["ObjectChild"]:
        """A generator over the dependents of the instance.

        In this context, dependents are child objects that the parent relies upon, rather
        than contains. For example, the Features stored under a Container are *not*
        dependents of the Container, but they are children.

        Returns
        -------
        Iterator[ObjectChild]
            The dependents, as `ObjectChild` instances.

        """
        for d in self._kml_dependents:
            d_obj = getattr(self, d.name, None)
            if d_obj is None:
                continue
            elif isinstance(d_obj, _ListObject):
                if not d_obj._yield_if_empty and len(d_obj) <= 0:
                    continue
                elif d_obj._yield_self:
                    yield ObjectChild(self, d_obj)
                else:
                    yield from (ObjectChild(self, dd) for dd in d_obj)
            elif isinstance(d_obj, Iterable):
                yield from (ObjectChild(self, dd) for dd in d_obj)
            else:
                yield ObjectChild(self, d_obj)

    @property
    def kml_tag(self) -> str:
        """The instance's KML tag string.

        Property that specifies the name of the XML tag that forms the root of
        the KML representation of this `_BaseObject`.
        """
        return self._kml_tag

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided `etree.Element`.

        Generate the KML representation of the internal fields of this `_BaseObject`, and
        append it to the provided root `etree.Element`.

        Parameters
        ----------
        root : etree.Element
            The `etree.Element` to which the KML published for the object is to be
            appended.
        with_children : bool, default = True
            True if the children of this instance should be included in the build.
        with_dependents : bool, default = True
            True if the dependents of this instance should be included in the build.

        """
        for f in (f for f in self._kml_fields if f.dumper != _NoDump):
            value = f.dumper.dump(getattr(self, f.name))
            if value is not None:
                etree.SubElement(root, with_ns(f.typename)).text = value
        if with_dependents:
            for dd in self.dependents:
                branch = dd.child.construct_kml(with_children, with_dependents)
                root.append(branch)
        if with_children:
            for dc in self.children:
                branch = dc.child.construct_kml(with_children, with_dependents)
                root.append(branch)

    def construct_kml(
        self, with_children: bool = True, with_dependents: bool = True
    ) -> etree.Element:
        """Construct this `_BaseObject`'s KML representation.

        Parameters
        ----------
        root : etree.Element
        with_children : bool, default = True
            True if the children of this instance should be included in the build.
        with_dependents : bool, default = True
            True if the dependents of this instance should be included in the build.

        Returns
        -------
        etree.Element
            The KML representation of the `_BaseObject` as an `etree.Element`.

        """
        # if the object is *not* in the "kml" namespace, ensure that the `id` that is
        # assigned *is* in the "kml" namespace, or GEP is likely to crash.
        attribs: dict[str, str] | None = None
        if not self._suppress_id:
            attribs = {}
            if ":" in self._kml_tag:
                attribs[with_ns("kml:id")] = self.id
            else:
                attribs["id"] = self.id

        root = etree.Element(_tag=with_ns(self.kml_tag), attrib=attribs)
        self.build_kml(root, with_children, with_dependents)
        return root

    def create_kml(self, root: etree.Element, parent: "_BaseObject") -> etree.Element:
        """Construct a complete `<Create>` element tag as a child of an `Update`.

        Parameters
        ----------
        root : etree.Element
            The etree.Element of the `<Update>` tag that will be appended to.
        parent : _BaseObject
            The immediate parent of this `_BaseObject`. Required for GEP synchronization,
            specifically the `targetId` attribute.

        Returns
        -------
        etree.Element
            The newly constructed child tag.

        """
        # if the parent is *not* in the "kml" namespace, ensure that the `targetId` that
        # is assigned *is* in the "kml" namespace, or GEP is likely to crash.
        parent_attribs = {}
        if ":" in parent._kml_tag:
            parent_attribs[with_ns("kml:targetId")] = parent.id
        else:
            parent_attribs["targetId"] = parent.id
        parent_element = etree.SubElement(
            root, with_ns(parent.kml_tag), attrib=parent_attribs
        )
        # if the child is *not* in the "kml" namespace, ensure that the `id` that is
        # assigned *is* in the "kml" namespace, or GEP is likely to crash.
        child_attribs: dict[str, str] | None = None
        if not self._suppress_id:
            child_attribs = {}
            if ":" in self._kml_tag:
                child_attribs[with_ns("kml:id")] = self.id
            else:
                child_attribs["id"] = self.id
        child_element = etree.SubElement(
            parent_element, with_ns(self.kml_tag), attrib=child_attribs
        )
        self.build_kml(child_element, False, True)
        return child_element

    def change_kml(self, root: etree.Element) -> None:
        """Construct a complete `<Change>` element tag as a child of an `Update`.

        Parameters
        ----------
        root : etree.Element
            The etree.Element of the `<Update>` tag that will be appended to.

        """
        # if the target is *not* in the "kml" namespace, ensure that the `targetId` that
        # is assigned *is* in the "kml" namespace, or GEP is likely to crash.
        attribs: dict[str, str] | None = None
        if not self._suppress_id:
            attribs = {}
            if ":" in self._kml_tag:
                attribs[with_ns("kml:targetId")] = self.id
            else:
                attribs["targetId"] = self.id
        item = etree.SubElement(root, _tag=with_ns(self.kml_tag), attrib=attribs)
        self.build_kml(item, with_children=False, with_dependents=False)

    def delete_kml(self, root: etree.Element) -> None:
        """Construct a complete `<Delete>` element tag as a child of an `Update`.

        Parameters
        ----------
        root : etree.Element
            The etree.Element of the `<Update>` tag that will be appended to.

        """
        # if the target is *not* in the "kml" namespace, ensure that the `targetId` that
        # is assigned *is* in the "kml" namespace, or GEP is likely to crash.
        attribs: dict[str, str] | None = None
        if not self._suppress_id:
            attribs = {}
            if ":" in self._kml_tag:
                attribs[with_ns("kml:targetId")] = self.id
            else:
                attribs["targetId"] = self.id
        etree.SubElement(root, _tag=with_ns(self.kml_tag), attrib=attribs)

    def activate(self, value: bool, cascade: bool = False) -> None:
        """Activate or deactivate this `_BaseObject` for display in GEP.

        Parameters
        ----------
        value : bool
            `True` to display the `_BaseObject`, `False` to remove it.
        cascade : bool, default = False
            `True` if the activation is to be cascaded to all child objects, `False`
            otherwise.

        """
        if self._state == ObjectState.CREATING:
            self._state = ObjectState.IDLE if not value else self._state
        elif self._state == ObjectState.CREATED:
            self._state = ObjectState.DELETE_CREATED if not value else self._state
        elif self._state == ObjectState.CHANGING:
            self._state = ObjectState.DELETE_CHANGED if not value else self._state
        elif self._state == ObjectState.DELETE_CREATED:
            self._state = ObjectState.CREATING if value else self._state
        elif self._state == ObjectState.DELETE_CHANGED:
            self._state = ObjectState.CHANGING if value else self._state
        else:  # implies default state is IDLE
            self._state = ObjectState.CREATING if value else self._state
        # cascade Activate downwards for children
        if value:
            for d in self.dependents:
                d.child.activate(True, cascade)
            for c in self.children:
                c.child.activate(True, cascade)
        else:
            for d in self.dependents:
                d.child.force_idle()
            for c in self.children:
                c.child.force_idle()

    def field_changed(self) -> None:
        """Flag that an attribute of the `_BaseObject` may have changed.

        If this method is called, it indicates that re-synchronization with GEP may be
        required for the `_BaseObject`.

        """
        if self._state == ObjectState.CREATED:
            self._state = ObjectState.CHANGING
        elif self._state == ObjectState.DELETE_CREATED:
            self._state = ObjectState.DELETE_CHANGED
        # if not already CREATED or DELETE_CREATED, change nothing

    def synchronized(self) -> None:
        """Flag that a synchronization update has been emitted for the `_BaseObject`.

        Calling this method modifies the `state` to the next appropriate value.

        """
        if self._state == ObjectState.CREATING:
            self._state = ObjectState.CREATED
            # force-cascade CREATED to dependents when the parent has been created,
            # because dependents are always created at the same time
            for d in self.dependents:
                d.child.force_created()
        elif self._state == ObjectState.CHANGING:
            self._state = ObjectState.CREATED
        elif (
            self._state == ObjectState.DELETE_CREATED
            or self._state == ObjectState.DELETE_CHANGED
        ):
            self._state = ObjectState.IDLE

    def force_created(self) -> None:
        """Force this `_BaseObject` and **all of its dependents** to the `CREATED` state.

        This is typically done after the parent object has been synchronized, i.e.
        created.
        """
        self._state = ObjectState.CREATED
        for d in self.dependents:
            d.child.force_created()

    def force_idle(self) -> None:
        """Force this `_BaseObject` and **all of its children AND dependents** to the `IDLE` state.

        This is typically done after the object has been deactivated, which will cause
        it to be deleted from GEP at the next synchronization update. When the
        `_BaseObject` is deleted by GEP, all of its children, and any `Feature`s that it
        encloses, will also be automatically deleted. There is no need to emit `<Delete>`
        tags for these deletions, and in fact doing so will likely cause GEP to have
        conniptions.

        """
        self._state = ObjectState.IDLE
        for d in self.dependents:
            d.child.force_idle()
        for c in self.children:
            c.child.force_idle()


class Object(_BaseObject, ABC):
    """A KML `<Object>` tag constructor.

    This is an abstract base class and cannot be used directly in a KML file. It provides
    the `id` attribute, which allows unique identification of a KML element, and the
    `targetId` attribute, which is used to reference objects that have already been
    loaded into Google Earth. The `id` attribute must be assigned if the <Update>
    mechanism is to be used.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#object

    """

    _suppress_id: bool = False
    _kml_children: tuple[_ChildDef, ...] = _BaseObject._kml_children

    def __init__(self) -> None:
        """Object instance constructor."""
        super().__init__()


class ObjectChild:
    """Describes a parent:child relationship between two `_BaseObject` instances.

    Parameters
    ----------
    parent: _BaseObject
    child: _BaseObject

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        parent: _BaseObject,
        child: _BaseObject,
    ) -> None:
        """ObjectChild instance constructor."""
        self.parent = parent
        self.child = child


_LOB = TypeVar("_LOB", bound="_BaseObject")


class _ListObject(_BaseObject, list[_LOB], Generic[_LOB]):
    """Internal class for KML objects that may contain a list of a type of child object.

    The most obvious example is `Container` and it's subclasses `Document` and `Folder`,
    but there are others: `MultiGeometry`, `MultiTrack`, `Schema` and `Tour`.

    The point to this is to allow for detection and handling of changes to the contents
    of these lists.
    """

    _yield_self: bool = False
    _yield_if_empty: bool = True

    def clear(self) -> None:
        """Override superclass `clear()` to call `field_changed()."""
        self.field_changed()
        if isinstance(self, _DeletableMixin):
            self._deleted.extend(self)
        super().clear()

    def remove(self, value: _LOB) -> None:
        """Override superclass `remove()` to call `field_changed()."""
        super().remove(value)
        self.field_changed()
        if isinstance(self, _DeletableMixin):
            self._deleted.append(value)

    def append(self, value: _LOB) -> None:
        """Override superclass `append()` to call `field_changed()."""
        self.field_changed()
        super().append(value)

    def extend(self, iterable: Iterable[_LOB]) -> None:
        """Override superclass `extend()` to call `field_changed()."""
        self.field_changed()
        super().extend(iterable)


class _DeletableMixin:
    """A mixin class for tracking deletion of child objects.

    Intended for use primarily with concrete `Container` implementations, but has other
    uses as well.
    """

    def __init__(self) -> None:
        self._deleted = list[_BaseObject]()



"""AbstractView module."""

from abc import ABC
from typing import Any, Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.errors import ViewerOptionInvalidError
from pyLiveKML.objects.Object import (
    _BaseObject,
    _ChildDef,
    _DependentDef,
    _ListObject,
    Object,
    ObjectChild,
)
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types import ViewerOptionEnum
from pyLiveKML.utils import with_ns


class ViewerOption(_BaseObject):
    """Enables or disables special viewing modes.

    Used only in conjunction with subclasses of :class:`pyLiveKML.objects.AbstractView`.

    Notes
    -----
    * Applies only to Google Earth 6.0 and later.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-abstractview.

    Parameters
    ----------
    option : ViewerOptionEnum
        The view option to be configured.
    enabled : bool
        Whether the view option is to be enabled or disabled.

    """

    _kml_tag = "gx:option"

    def __init__(self, option: ViewerOptionEnum, enabled: bool):
        """GxViewerOption instance constructor."""
        super().__init__()
        self.option = option
        self.enabled = enabled

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element."""
        root.attrib["name"] = self.option.value
        root.attrib["enabled"] = str(int(self.enabled))


class _ViewerOptions(_ListObject[ViewerOption], _BaseObject):
    """A KML `<gx:ViewerOptions>` tag constructor.

    Private class to manage viewer options in `AbstractView` subclasses.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-abstractview

    Parameters
    ----------
    disallowed_view_options : tuple[ViewerOptionEnum, ...]
        The viewer options that are not permitted for the owning concrete subclass of
        `AbstractView`.
    items : Iterable[ViewerOption] | ViewerOption | None, default = None
        The viewer options that are to be configured.

    """

    _kml_tag = "gx:ViewerOptions"
    _kml_dependents = _BaseObject._kml_dependents + (_DependentDef("items"),)
    _yield_self = True
    _yield_if_empty = False

    def __init__(
        self,
        disallowed_view_options: tuple[ViewerOptionEnum, ...],
        items: Iterable[ViewerOption] | ViewerOption | None = None,
    ) -> None:
        """ViewerOptions instance constructor."""
        super().__init__()
        self.disallowed_view_options = disallowed_view_options
        self.items = items

    @property
    def items(self) -> Iterator[ViewerOption]:
        yield from self

    @items.setter
    def items(self, value: Iterable[ViewerOption] | ViewerOption | None) -> None:
        self.clear()
        if value is not None:
            if isinstance(value, ViewerOption):
                if value.option in self.disallowed_view_options:
                    raise ViewerOptionInvalidError(value.option.value)
                self.append(value)
            else:
                final = {v.option: v.enabled for v in value}
                errors = [
                    k.value for k in final.keys() if k in self.disallowed_view_options
                ]
                if errors:
                    raise ViewerOptionInvalidError(errors)
                self.extend(ViewerOption(k, v) for k, v in final.items())


class AbstractView(Object, ABC):
    """A KML `<AbstractView>` tag constructor.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#abstractview

    Parameters
    ----------
    viewer_options : ViewerOption | Iterable[ViewerOption] | None, default = None
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None, default = None
        `TimeStamp` or `TimeSpan` assigned to the object.

    Attributes
    ----------
    viewer_options : list[ViewerOption]
    time_primitive : TimePrimitive | None

    Raises
    ------
    ViewerOptionInvalidError
        If one of the viewer options assigned is not permitted for the concrete subclass.

    """

    _kml_children = Object._kml_children + (_ChildDef("time_primitive"),)
    _kml_dependents = Object._kml_dependents + (_DependentDef("viewer_options"),)
    _disallowed_view_options: tuple[ViewerOptionEnum, ...] = tuple()

    def __init__(
        self,
        viewer_options: Iterable[ViewerOption] | ViewerOption | None = None,
        time_primitive: TimePrimitive | None = None,
    ) -> None:
        """AbstractView instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)
        self.viewer_options = _ViewerOptions(
            self._disallowed_view_options, viewer_options
        )
        self.time_primitive = time_primitive

    def __setattr__(self, name: str, value: Any) -> None:
        """Manipulate attribute assignments.

        Ensures that if a time primitive is assigned to an `AbstractView`, it's
        `_kml_tag` is prefixed with the "gx:" namespace.
        """
        if name == "time_primitive" and isinstance(value, TimePrimitive):
            value._kml_tag = f"gx:{type(value)._kml_tag}"
        return super().__setattr__(name, value)



"""TimeSpan module."""

from datetime import datetime

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef, _DateTimeParse
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class TimeSpan(TimePrimitive):
    """A KML `<TimeSpan>` tag constructor.

    Represents an extent in time bounded by begin and end `datetime`s.

    Notes
    -----
    If `begin` or `end` is `None`, then that end of the period is unbounded.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#style

    Parameters
    ----------
    begin: datetime | str | None, default = None
    end: datetime | str | None, default = None

    Attributes
    ----------
    begin: datetime | None
    end: datetime | None

    """

    _kml_tag = "TimeSpan"
    _kml_fields = TimePrimitive._kml_fields + (
        _FieldDef("begin", parser=_DateTimeParse),
        _FieldDef("end", parser=_DateTimeParse),
    )

    def __init__(
        self,
        begin: datetime | str | None = None,
        end: datetime | str | None = None,
    ):
        """TimeSpan instance constructor."""
        TimePrimitive.__init__(self)
        self.begin = begin
        self.end = end



"""TimePrimitive module."""

from abc import ABC

from pyLiveKML.objects.Object import Object


class TimePrimitive(Object, ABC):
    """A KML `<TimePrimitive>` tag constructor.

    This is an abstract class and cannot be used directly in a KML file. Extended by the
    `TimeSpan` and `TimeStamp` (and `GxTimeSpan` and `GxTimeStamp`) classes.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    def __init__(self) -> None:
        """TimePrimitive instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)



"""LookAt module."""

from typing import Any, Iterable, cast

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView, ViewerOption
from pyLiveKML.objects.Object import (
    _FieldDef,
    _Angle180,
    _Angle360,
    _Angle90,
    _AnglePos90,
    _NoParse,
)
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types import AltitudeModeEnum, GeoCoordinates, ViewerOptionEnum


class LookAt(AbstractView):
    """A KML `<LookAt>` tag constructor.

    Defines a virtual camera that is associated with any element derived from `Feature`.
    The `LookAt` element positions the "camera" in relation to the object that is being
    viewed. In Google Earth, the view "flies to" this `LookAt` viewpoint when the user
    double-clicks an item in the "Places" panel or double-clicks an icon in the 3D
    viewer.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#lookat

    Parameters
    ----------
    lla : GeoCoordinates | tuple[float, float, float|None] | tuple[float, float] | None, default = None
        The longitude, latitude and altitude (in that order, in decimal degrees) of the
        `LookAt` target. If `None`, `lon` and `lat` must be supplied or an exception will
        be thrown.
    heading : float, default = 0
        The heading (in decimal degrees) of the `LookAt` facing.
    tilt : float, default = 0
        The tilt (in decimal degrees) of the `LookAt` facing.
    range : float, default = 0
        Distance in meters from the point specified by `longitude`, `latitude` and
        `altitude` to the `LookAt` position.
    altitude_mode : AltitudeModeEnum | None, default = None
    lon : float | None, default = None
        The longitude (in decimal degrees) of the the `LookAt` target. If `None`, `lla`
        must be supplied.
    lat : float | None, default = None
        The latitude (in decimal degrees) of the the `LookAt` target. If `None`, `lla`
        must be supplied.
    alt : float | None, default = None
        The altitude (in metres, with respect to `altitude_mode`) of the the `LookAt`
        target.
    viewer_options : ViewerOption | Iterable[ViewerOption] | None, default = None
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None, default = None
        `TimeStamp` or `TimeSpan` assigned to the object.

    Attributes
    ----------
    viewer_options : list[ViewerOption]
    time_primitive : TimePrimitive | None
    lla : tuple[float, float, float]
        The longitude, latitude (respectively, in decimal degrees) and altitude (in
        metres) of the `LookAt` target.
    lon : float
        The longitude of the `LookAt` target, in decimal degrees.
    lat : float
        The latitude of the `LookAt` target, in decimal degrees.
    alt : float
        The altitude of the `LookAt` target, in metres, with respect to `altitude_mode`.
    heading : float
        The heading of the `LookAt` facing, in decimal degrees.
    tilt : float
        The tilt of the `LookAt` facing, in decimal degrees.
    range : float
        Distance in meters from the `LookAt` position to the target point specified by
        `lla`.
    altitude_mode : AltitudeModeEnum | None

    Raises
    ------
    ValueError
        If `lla` is None, and either `lat` or `lon` is also `None`, a `ValueError` will
        be raised by the constructor.
    ViewerOptionInvalidError
        If the `GROUND_NAVIGATION` viewer option is assigned.

    """

    _kml_tag = "LookAt"
    _kml_fields = AbstractView._kml_fields + (
        _FieldDef("lon", "longitude", parser=_Angle180),
        _FieldDef("lat", "latitude", parser=_Angle90),
        _FieldDef("alt", "altitude"),
        _FieldDef("heading", parser=_Angle360),
        _FieldDef("tilt", parser=_AnglePos90),
        _FieldDef("range", parser=_NoParse),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )
    _disallowed_view_options = AbstractView._disallowed_view_options + (
        ViewerOptionEnum.GROUND_NAVIGATION,
    )

    def __init__(
        self,
        lla: (
            GeoCoordinates
            | tuple[float, float, float | None]
            | tuple[float, float]
            | None
        ) = None,
        heading: float = 0,
        tilt: float = 0,
        range: float = 0,
        altitude_mode: AltitudeModeEnum | None = None,
        lon: float | None = None,
        lat: float | None = None,
        alt: float | None = None,
        viewer_options: ViewerOption | Iterable[ViewerOption] | None = None,
        time_primitive: TimePrimitive | None = None,
    ):
        """LookAt instance constructor."""
        AbstractView.__init__(self, viewer_options, time_primitive)
        self.lon: float
        self.lat: float
        self.alt: float
        if lla is None and (lon is None or lat is None):
            raise ValueError(
                "You must supply either `lla` or `lon` and `lat` (and optionally `alt`)."
            )
        if lla is None:
            self.lon = cast(float, lon)
            self.lat = cast(float, lat)
            self.alt = cast(float, alt)
        else:
            self.lla = lla
        self.heading = heading
        self.tilt = tilt
        self.range = range
        self.altitude_mode = altitude_mode

    @property
    def lla(self) -> tuple[float, float, float]:
        """Get or set the longitude, latitude and altitude as a tuple.

        Parameters
        ----------
        value : GeoCoordinates | tuple[float, float, float|None] | tuple[float, float])
            The longitude, latitude and altitude as a tuple with optional altitude, or as
            `GeoCoordinates`.

        Returns
        -------
        tuple[float, float, float]
            The longitude, latitude and altitude as a tuple of floats.

        """
        return (self.lon, self.lat, self.alt)

    @lla.setter
    def lla(
        self,
        value: GeoCoordinates | tuple[float, float, float | None] | tuple[float, float],
    ) -> None:
        if isinstance(value, GeoCoordinates):
            v = value.values
            self.lon, self.lat = v[:2]
            self.alt = 0 if v[2] is None else v[2]
        elif len(value) >= 3:
            v = value[:3]
            self.lon, self.lat = v[:2]
            self.alt = 0 if v[2] is None else v[2]
        else:
            self.lon, self.lat = value
            self.alt = 0



"""SubStyle module."""

from abc import ABC

from pyLiveKML.objects.Object import _FieldDef, Object


class SubStyle(Object, ABC):
    """A KML `<SubStyle>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file.

    Notes
    -----
    There is no description of the `SubStyle` class in the Google KML documentation,
    although it is include in the inheritance tree at the top of the page.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference.

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    _kml_fields: tuple[_FieldDef, ...] = tuple()

    def __init__(self) -> None:
        """SubStyle instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)



"""MultiGeometry module."""

from typing import Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Object import _ChildDef, _DeletableMixin, _ListObject


class MultiGeometry(_DeletableMixin, _ListObject[Geometry], Geometry):
    """A KML `<MultiGeometry>` tag constructor.

    A container for zero or more geometry primitives associated with the same feature.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#multigeometry

    Parameters
    ----------
    contents : Geometry | Iterable[Geometry] | None, default = None
        The `Geometry` instances to be hosted under this `MultiGeometry`.

    Attributes
    ----------
    Nil

    """

    _kml_tag = "MultiGeometry"
    _kml_children = Geometry._kml_children + (_ChildDef("contents"),)
    _yield_self = True

    def __init__(self, contents: Geometry | Iterable[Geometry] | None = None):
        """MultiGeometry instance constructor."""
        _DeletableMixin.__init__(self)
        _ListObject[Geometry].__init__(self)
        Geometry.__init__(self)
        self.contents = contents

    @property
    def contents(self) -> Iterator[Geometry]:
        """Retrieve a generator over the `Geometry`s in this `MultiGeometry`.

        If the property setter is called, replaces the current list of contained
        `Geometry`s with those provided.

        Parameters
        ----------
        value : Geometry | Iterable[Geometry] | None
            The new `Geometry` elements for the `MultiGeometry`.

        :returns: A generator over the `Geometry`s in the `MultiGeometry`.
        :rtype: Iterator[Geometry]

        """
        yield from self

    @contents.setter
    def contents(self, value: Geometry | Iterable[Geometry] | None) -> None:
        self.clear()
        if value is not None:
            if isinstance(value, Geometry):
                self.append(value)
            else:
                self.extend(value)



"""PolyStyle module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.ColorStyle import ColorStyle
from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.types import ColorModeEnum, GeoColor


class PolyStyle(ColorStyle):
    """A KML `<PolyStyle>` tag constructor.

    Specifies the drawing style for all polygons, including polygon extrusions (which
    look like the walls of buildings) and line extrusions (which look like solid fences).

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#polystyle

    Parameters
    ----------
    color: GeoColor | int | None, default = None
    color_mode: ColorModeEnum | None, default = None
    fill: bool | None, default = None
        Specifies whether to fill the polygon.
    outline: bool | None, default = None
        Specifies whether to outline the polygon. `Polygon` outlines use the current
        `LineStyle`.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "PolyStyle"
    _kml_fields = ColorStyle._kml_fields + (
        _FieldDef("fill"),
        _FieldDef("outline"),
    )

    def __init__(
        self,
        color: GeoColor | int | None = None,
        color_mode: ColorModeEnum | None = None,
        fill: bool | None = None,
        outline: bool | None = None,
    ):
        """PolyStyle instance constructor."""
        ColorStyle.__init__(self, color, color_mode)
        self.fill = fill
        self.outline = outline



"""BalloonStyle module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef, _ColorParse
from pyLiveKML.objects.SubStyle import SubStyle
from pyLiveKML.types import DisplayModeEnum, GeoColor


class BalloonStyle(SubStyle):
    """A KML `<BalloonStyle>` tag constructor.

    Specifies how the description balloon for placemarks is drawn. The `bgColor`, if
    specified, is used as the background color of the balloon.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#balloonstyle

    Parameters
    ----------
    bg_color : GeoColor | int | None, default = None
        Background color of the balloon.
    text_color : GeoColor | int | None, default = None
    text : str | None, default = None
        Text displayed in the balloon. If no text is specified, Google Earth draws the
        default balloon (with the `Feature` `<name>` in boldface, the `Feature`
        `<description>`, links for driving directions, a white background, and a tail
        that is attached to the point coordinates of the `Feature`, if specified).
    display_mode : DisplayModeEnum | None, default = None
        If `<displayMode>` is "default", Google Earth uses the information supplied in
        `<text>` to create a balloon. If `<displayMode>` is "hide", Google Earth does not
        display the balloon. In Google Earth, clicking the List View icon for a
        `Placemark` whose balloon's `<displayMode>` is "hide" causes Google Earth to fly
        to the `Placemark`.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "BalloonStyle"
    _kml_fields = SubStyle._kml_fields + (
        _FieldDef("bg_color", "bgColor", _ColorParse),
        _FieldDef("text_color", "textColor", _ColorParse),
        _FieldDef("text"),
        _FieldDef("display_mode", "displayMode"),
    )

    def __init__(
        self,
        bg_color: GeoColor | int | None = None,
        text_color: GeoColor | int | None = None,
        text: str | None = None,
        display_mode: DisplayModeEnum | None = None,
    ):
        """BalloonStyle instance constructor."""
        SubStyle.__init__(self)
        self.bg_color = bg_color
        self.text_color = text_color
        self.text = text
        self.display_mode = display_mode



"""Tour module."""

from typing import Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Feature import Feature
from pyLiveKML.objects.Object import (
    _BaseObject,
    _ChildDef,
    _DeletableMixin,
    _DependentDef,
    _FieldDef,
    _ListObject,
    Object,
)
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.objects.TourPrimitive import TourPrimitive


class Playlist(_DeletableMixin, _ListObject[TourPrimitive], _BaseObject):
    """A KML `<Playlist>` tag constructor.

    Contains any number of `TourPrimitive` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#contains_6

    Parameters
    ----------
    items : TourPrimitive | Iterable[TourPrimitive] | None, default = None
        The `TourPrimitive` items in the playlist.

    Attributes
    ----------
    Nil

    """

    _kml_tag = "gx:Playlist"
    _kml_children = _BaseObject._kml_children + (_ChildDef("items"),)
    _yield_self = True

    def __init__(
        self, items: TourPrimitive | Iterable[TourPrimitive] | None = None
    ) -> None:
        """Playlist instance constructor."""
        _DeletableMixin.__init__(self)
        _ListObject[TourPrimitive].__init__(self)
        _BaseObject.__init__(self)
        self.items = items

    @property
    def items(self) -> Iterator[TourPrimitive]:
        """Retrieve a generator over the `TourPrimitive`s in this `Tour`.

        If the property setter is called, replaces the current list of contained
        `TourPrimitive`s with those provided.

        Parameters
        ----------
        value : TourPrimitive | Iterable[TourPrimitive] | None
            The new `TourPrimitive` elements for the `Tour`.

        :returns: A generator over the `TourPrimitive`s in the `Tour`.
        :rtype: Iterator[TourPrimitive]

        """
        yield from self

    @items.setter
    def items(self, value: TourPrimitive | Iterable[TourPrimitive] | None) -> None:
        self.clear()
        if value is not None:
            if isinstance(value, TourPrimitive):
                self.append(value)
            else:
                self.extend(value)


class Tour(Feature):
    """A KML `<gx:Tour>` tag constructor.

    A `Tour` contains a single `PlayList`, which in turn contains an ordered list of
    `TourPrimitive` elements that define a tour in any KML browser.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxtour
    * https://developers.google.com/kml/documentation/touring

    Parameters
    ----------
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.
    playlist: TourPrimitive | list[TourPrimitive] | None, default = None

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:Tour"
    _kml_dependents = Object._kml_dependents + (_DependentDef("playlist"),)

    def __init__(
        self,
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
        playlist: TourPrimitive | Iterable[TourPrimitive] | None = None,
    ) -> None:
        """Tour instance constructor."""
        super().__init__(
            name=name,
            visibility=visibility,
            is_open=is_open,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
        )
        self.playlist = Playlist(playlist)



"""ScreenOverlay module."""

from abc import ABC
from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Icon import Icon
from pyLiveKML.objects.Object import _ChildDef, _FieldDef
from pyLiveKML.objects.Overlay import Overlay
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types.GeoColor import GeoColor
from pyLiveKML.types.Vec2 import OverlayXY, RotationXY, ScreenXY, Size


class ScreenOverlay(Overlay):
    """A KML `<ScreenOverlay>` tag constructor.

    Draws an image overlay fixed to the screen. Sample uses for `ScreenOverlay`s are
    compasses, logos, and heads-up displays. `ScreenOverlay` sizing is determined by the
    `size` attribute. Positioning of the overlay is handled by mapping a point in the
    image specified by `overlay_xy` to a point on the screen specified by `screen_xy`.
    Then the image is rotated by `rotation` degrees about a point relative to the screen
    specified by `rotation_xy`.

    Notes
    -----
    `icon` specifies the image to be used as the overlay. This file can be either on a
    local file system or on a web server. If `icon` is `None`, a rectangle is drawn using
    the color and size defined by the screen overlay.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#screenoverlay.

    Parameters
    ----------
    overlay_xy: OverlayXY
        Specifies a point on (or outside of) the overlay image that is mapped to the
        screen coordinate `screen_xy`. It requires `x` and `y` values, and the units for
        those values.
    screen_xy: ScreenXY
        Specifies a point relative to the screen origin that the overlay image is mapped
        to.
    rotation_xy: RotationXY
        Point relative to the screen about which the screen overlay is rotated.
    size: Size
        Specifies the size of the image for the screen overlay.
    rotation: float, default = 0
        Indicates the angle of rotation of the parent object. A value of 0 means no
        rotation. The value is an angle in degrees counterclockwise starting from north.
        Use ±180 to indicate the rotation of the parent object from 0. The center of the
        `rotation`, if not (.5,.5), is specified in `rotation_xy`.
    icon : str | Icon | None, default = None
        Defines the location of the image to be used as the `Overlay`. This location can
        be either on a local file system or on a web server. If this attribute is `None`,
        a rectangle is drawn using the color and size defined by the ground or screen
        overlay. If a simple `str` is supplied, then it will be used as the `href` of an
        `Icon`.
    color : GeoColor | int | None, default = None
        The image or rectangle color and transparency.
    draw_order : int | None, default = None
        Defines the stacking order for the images in overlapping `Overlay`s. `Overlay`s
        with higher `draw_order` values are drawn on top of `Overlay`s with lower
        `draw_order` values.
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.

        A given KML `Feature` can contain a combination of these types of custom data.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "ScreenOverlay"
    _kml_fields = Overlay._kml_fields + (_FieldDef("rotation"),)
    _kml_children = Overlay._kml_children + (
        _ChildDef("overlay_xy"),
        _ChildDef("screen_xy"),
        _ChildDef("rotation_xy"),
        _ChildDef("size"),
    )

    def __init__(
        self,
        # ScreenOverlay parameters
        overlay_xy: OverlayXY | None = None,
        screen_xy: ScreenXY | None = None,
        rotation_xy: RotationXY | None = None,
        size: Size | None = None,
        rotation: float = 0,
        # Overlay parameters
        icon: str | Icon | None = None,
        draw_order: int | None = None,
        color: GeoColor | int | None = None,
        name: str | None = None,
        visibility: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
    ):
        """IconStyle instance constructor."""
        Overlay.__init__(
            self,
            icon=icon,
            draw_order=draw_order,
            color=color,
            name=name,
            visibility=visibility,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
        )
        ABC.__init__(self)
        self.overlay_xy = overlay_xy
        self.screen_xy = screen_xy
        self.rotation_xy = rotation_xy
        self.size = size
        self.rotation = rotation



"""LineString module."""

from typing import Any, Iterable, Iterator, cast

from lxml import etree  # type: ignore

from pyLiveKML.errors import LineStringCoordsError
from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Object import _FieldDef, _KMLDump, _KMLParser
from pyLiveKML.types import AltitudeModeEnum, GeoCoordinates


class _CoordsParser(_KMLParser):

    @classmethod
    def parse(cls, value: Any) -> Any:
        if isinstance(next(iter(value)), GeoCoordinates):
            result = value
        else:
            result = tuple((GeoCoordinates(*c) for c in value))
        if len(result) < 2:
            raise LineStringCoordsError(
                "There must be at least two points in the line."
            )
        return result


class _CoordsDumper(_KMLDump):

    @classmethod
    def dump(cls, value: Any) -> Any:
        def _build() -> Iterable[str]:
            yield from (str(c) for c in cast(Iterable[GeoCoordinates], value))

        return " ".join(_build())


class LineString(Geometry):
    """A KML `<LineString>` geometry tag constructor.

    Defines a connected set of line segments. Use a `LineStyle` to specify the color,
    color mode, and width of the line. When a `LineString` is extruded, the line is
    extended to the ground, forming a polygon that looks somewhat like a wall or
    fence. For extruded `LineString`s, the line itself uses the current `LineStyle`, and
    the extrusion uses the current `PolyStyle`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#linestring.

    Parameters
    ----------
    coordinates : Iterable[GeoCoordinates] | Iterable[tuple[float, float, float]] | Iterable[tuple[float, float]]
        The coordinates of the vertices of the `LinearRing`. Note that there is no need
        to append a final joining coordinate, such that c[0] == c[-1]; c[0] will be
        appended automatically when the coordinates are published.
    altitude_mode : AltitudeModeEnum | None, default = None
        Specifies how altitude components in the `coordinates` attribute are interpreted.
    extrude : bool | None, default = None
        Specifies whether to connect the `LinearRing` to the ground. To extrude this
        geometry, `altitude_mode` must be one of RELATIVE_TO_GROUND,
        RELATIVE_TO_SEAFLOOR or ABSOLUTE. Only the vertices of the `LinearRing` are
        extruded, not the center of the geometry. The vertices are extruded toward
        the center of the Earth's geoid.
    tessellate : bool | None, default = None
        Specifies whether to allow the `LinearRing` to follow the terrain. To enable
        tessellation, `altitude_mode` must be CLAMP_TO_GROUND or CLAMP_TO_SEAFLOOR.
        Very large `LinearRing`s should enable tessellation so that they follow the
        curvature of the earth; otherwise, they may go underground and be hidden.
    altitude_offset : float | None, default = None
        Modifies how the altitude values are rendered. This offset allows you to move an
        entire `LinearRing` up or down as a unit without modifying all the individual
        coordinate values that make up the `LinearRing`. (Although the `LinearRing` is
        displayed using the altitude offset value, the original altitude values are
        preserved in the KML file). Units are in meters.
    draw_order : int, default = 0
        An integer value that specifies the order for drawing multiple `LineString`s.
        `LineString`s drawn first may be partially or fully obscured by `LineString`s
        with a later draw order. This element may be required in conjunction with the
        `outer_color` and `outer_width` attributes in `LineStyle` when dual-colored lines
        cross each other.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "LineString"
    _kml_fields = Geometry._kml_fields + (
        _FieldDef("altitude_mode", "gx:altitudeMode"),
        _FieldDef("extrude"),
        _FieldDef("tessellate"),
        _FieldDef("altitude_offset", "gx:altitudeOffset"),
        _FieldDef("draw_order", "gx:drawOrder"),
        _FieldDef("coordinates", parser=_CoordsParser, dumper=_CoordsDumper),
    )

    def __init__(
        self,
        coordinates: (
            Iterable[GeoCoordinates]
            | Iterable[tuple[float, float, float]]
            | Iterable[tuple[float, float]]
        ),
        altitude_mode: AltitudeModeEnum | None = None,
        extrude: bool | None = None,
        tessellate: bool | None = None,
        altitude_offset: float | None = None,
        draw_order: int = 0,
    ):
        """LineString instance constructor."""
        Geometry.__init__(self)
        self.altitude_offset = altitude_offset
        self.extrude = extrude
        self.tessellate = tessellate
        self.altitude_mode = altitude_mode
        self.draw_order = draw_order
        self._coordinates = list[GeoCoordinates]()
        self.coordinates = coordinates

    @property
    def coordinates(self) -> Iterator[GeoCoordinates]:
        """Retrieve a generator over the `GeoCoordinates` of this `LineString`.

        If the property setter is called, replaces the current list of coordinates with
        those provided.

        Parameters
        ----------
        value : Iterable[GeoCoordinates] | Iterable[tuple[float, float, float]] | Iterable[tuple[float, float]]
            The new coordinates for the `LineString`.

        :returns: A generator over the `GeoCoordinates` of the `LineString`.
        :rtype: Iterator[GeoCoordinate]
        :raises: LineStringCoordsError
            If less than 2 coordinate values are supplied.

        """
        yield from self._coordinates

    @coordinates.setter
    def coordinates(
        self,
        value: (
            Iterable[GeoCoordinates]
            | Iterable[tuple[float, float, float]]
            | Iterable[tuple[float, float]]
        ),
    ) -> None:
        self._coordinates.clear()
        self._coordinates.extend(cast(Iterable[GeoCoordinates], value))



"""Point module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.types import AltitudeModeEnum, GeoCoordinates


class Point(Geometry):
    """A KML `<Point>` tag constructor.

    A geographic location defined by longitude, latitude, and (optional) altitude. When a
    `Point` is contained by a `Placemark`, the point itself determines the position of
    the `Placemark`'s name and icon. When a `Point` is extruded, it is connected to the
    ground with a line. This "tether" uses the current `LineStyle`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#point

    Parameters
    ----------
    coordinates : GeoCoordinates | tuple[float, float, float] | tuple[float, float]
        A single tuple consisting of floating point values for longitude, latitude, and
        altitude (in that order). Longitude and latitude values are in decimal degrees
        and are constrained to their circular or half-circular limit respectively.
    altitude_mode : AltitudeModeEnum | None, default = None
        Specifies how altitude components in `coordinates` are interpreted.
    extrude : bool | None, default = None
        Specifies whether to connect the `Point` to the ground with a line. To extrude a
        `Point`, the value for `altitude_mode` must be one of `RELATIVE_TO_GROUND`,
        `RELATIVE_TO_SEAFLOOR` or `ABSOLUTE`. The point is extruded toward the center of
        the Earth's geoid.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Point"
    _kml_fields = Geometry._kml_fields + (
        _FieldDef("coordinates"),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
        _FieldDef("extrude"),
    )

    def __init__(
        self,
        coordinates: GeoCoordinates | tuple[float, float, float] | tuple[float, float],
        altitude_mode: AltitudeModeEnum | None = None,
        extrude: bool | None = None,
    ):
        """Point instance constructor."""
        Geometry.__init__(self)
        self._coordinates: GeoCoordinates
        self.coordinates = coordinates
        self.extrude = extrude
        self.altitude_mode = altitude_mode

    @property
    def coordinates(self) -> GeoCoordinates:
        """Get or set the current `coordinates` of the `Point`.

        Parameters
        ----------
        value : GeoCoordinates | tuple[float, float, float] | tuple[float, float]
            The new coordinates value, expressable as a 2- or 3-tuple of floats for
            convenience.

        Returns
        -------
        GeoCoordinates
            The current `coordinates`, as a `GeoCoordinates` instance.

        """
        return self._coordinates

    @coordinates.setter
    def coordinates(
        self, value: GeoCoordinates | tuple[float, float, float] | tuple[float, float]
    ) -> None:
        if isinstance(value, GeoCoordinates):
            self._coordinates = value
        else:
            self._coordinates = GeoCoordinates(*value)



"""PhotoOverlay module."""

from abc import ABC
from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Icon import Icon
from pyLiveKML.objects.Object import (
    _BaseObject,
    _ChildDef,
    _FieldDef,
    _Angle180,
    _Angle90,
)
from pyLiveKML.objects.Overlay import Overlay
from pyLiveKML.objects.Point import Point
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types import GeoColor, GridOriginEnum, OverlayShapeEnum


class ViewVolume(_BaseObject):
    """A KML `<ViewVolume>` tag constructor.

    Defines how much of the current scene is visible. Specifying the field of view is
    analogous to specifying the lens opening in a physical camera. A small field of view,
    like a telephoto lens, focuses on a small part of the scene. A large field of view,
    like a wide-angle lens, focuses on a large part of the scene.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-photooverlay

    Parameters
    ----------
    left_fov: float, default = 0
        Angle, in degrees, between the camera's viewing direction and the left side of
        the view volume.
    right_fov: float, default = 0
        Angle, in degrees, between the camera's viewing direction and the right side of
        the view volume.
    bottom_fov: float, default = 0
        Angle, in degrees, between the camera's viewing direction and the bottom side of
        the view volume.
    top_fov: float, default = 0
        Angle, in degrees, between the camera's viewing direction and the top side of the
        view volume.
    near: float, default = 0
        Measurement in meters along the viewing direction from the camera viewpoint to
        the PhotoOverlay shape.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "ViewVolume"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("left_fov", "leftFov", _Angle180),
        _FieldDef("right_fov", "rightFov", _Angle180),
        _FieldDef("bottom_fov", "bottomFov", _Angle90),
        _FieldDef("top_fov", "topFov", _Angle90),
        _FieldDef("near"),
    )

    def __init__(
        self,
        left_fov: float = 0,
        right_fov: float = 0,
        bottom_fov: float = 0,
        top_fov: float = 0,
        near: float = 0,
    ) -> None:
        """_PhotoOverlay_ViewVolume instance constructor."""
        super().__init__()
        self.left_fov = left_fov
        self.right_fov = right_fov
        self.bottom_fov = bottom_fov
        self.top_fov = top_fov
        self.near = near


class ImagePyramid(_BaseObject):
    """A KML `<ImagePyramid>` tag constructor.

    A hierarchical set of images, each of which is an increasingly lower resolution
    version of the original image.

    Each image in the pyramid is subdivided into tiles, so that only the portions in view
    need to be loaded. Google Earth calculates the current viewpoint and loads the tiles
    that are appropriate to the user's distance from the image. As the viewpoint moves
    closer to the `PhotoOverlay`, Google Earth loads higher resolution tiles. Since all
    the pixels in the original image can't be viewed on the screen at once, this
    preprocessing allows Google Earth to achieve maximum performance because it loads
    only the portions of the image that are in view, and only the pixel details that can
    be discerned by the user at the current viewpoint.

    Notes
    -----
    * When you specify an `ImagePyramid` for a `PhotoOverlay`, you must also modify the
    `icon` attribute to include specifications for which tiles to load.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-photooverlay

    Parameters
    ----------
    tile_size : int, default = 256
        Size of the tiles, in pixels. Tiles must be square, and `tile_size` must be a
        power of 2. A tile size of 256 (the default) or 512 is recommended. The original
        image is divided into tiles of this size, at varying resolutions.
    max_width : int, default = 0
        Width in pixels of the original image.
    max_height : int, default = 0
        Height in pixels of the original image.
    grid_origin : GridOriginEnum, default = GridOriginEnum.LOWER_LEFT
        Specifies where to begin numbering the tiles in each layer of the pyramid. A
        value of `LOWER_LEFT` specifies that row 1, column 1 of each layer is in the
        bottom left corner of the grid.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "ImagePyramid"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("tile_size", "tileSize"),
        _FieldDef("max_width", "maxWidth"),
        _FieldDef("max_height", "maxHeight"),
        _FieldDef("grid_origin", "gridOrigin"),
    )

    def __init__(
        self,
        tile_size: int = 256,
        max_width: int = 0,
        max_height: int = 0,
        grid_origin: GridOriginEnum = GridOriginEnum.LOWER_LEFT,
    ) -> None:
        """_PhotoOverlay_ImagePyramid instance constructor."""
        super().__init__()
        self.tile_size = tile_size
        self.max_width = max_width
        self.max_height = max_height
        self.grid_origin = grid_origin


class PhotoOverlay(Overlay):
    """A KML `<PhotoOverlay>` tag constructor.

    `PhotoOverlay` allows you to geographically locate a photograph on the Earth and to
    specify viewing parameters for this `PhotoOverlay`. The `PhotoOverlay` can be a
    simple 2D rectangle, a partial or full cylinder, or a sphere (for spherical
    panoramas). The overlay is placed at the specified location and oriented toward the
    viewpoint.

    Because `PhotoOverlay` is derived from `Feature`, it can contain a concrete instance
    of `AbstractView` - either `Camera` or `LookAt`. The `Camera` (or `LookAt`) specifies
    a viewpoint and a viewing direction (also referred to as a view vector). The
    `PhotoOverlay` is positioned in relation to the viewpoint. Specifically, the plane of
    a 2D rectangular image is orthogonal (at right angles to) the view vector. The normal
    of this plane - that is, its front, which is the part with the photo - is oriented
    toward the viewpoint.

    The URL for the `PhotoOverlay` image is specified in the `icon` attribute, which is
    inherited from `Overlay`. The `icon` must specify the image file to use for the
    `PhotoOverlay`. In the case of a very large image, `icon` is a special URL that indexes
    into a pyramid of images of varying resolutions (see `ImagePyramid`).

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#photooverlay
    * https://developers.google.com/kml/documentation/photos

    Parameters
    ----------
    point : Point
        The `Point` acts as a `Point` inside a `Placemark` element. It draws an icon to
        mark the position of the `PhotoOverlay`. The icon drawn is specified by the
        `style_url` or `StyleSelector` fields, just as it is for `Placemark`.
    shape : OverlayShapeEnum | None, default = None
        The `PhotoOverlay` is projected onto the `shape`.
    rotation : float | None, default = None
        Adjusts how the photo is placed inside the field of view. This is useful if your
        photo has been rotated and deviates slightly from a desired horizontal view.
    view_volume : ViewVolume | None, default = None
        Defines how much of the current scene is visible.
    image_pyramid : ImagePyramid | None, default = None
        For very large images, you'll need to construct an image pyramid, which is a
        hierarchical set of images, each of which is an increasingly lower resolution
        version of the original image.
    icon : str | Icon | None, default = None
        Defines the location of the image to be used as the `Overlay`. This location can
        be either on a local file system or on a web server. If this attribute is `None`,
        a rectangle is drawn using the color and size defined by the ground or screen
        overlay. If a simple `str` is supplied, then it will be used as the `href` of an
        `Icon`.
    color : GeoColor | int | None, default = None
        The image or rectangle color and transparency.
    draw_order : int | None, default = None
        Defines the stacking order for the images in overlapping `Overlay`s. `Overlay`s
        with higher `draw_order` values are drawn on top of `Overlay`s with lower
        `draw_order` values.
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.

        A given KML `Feature` can contain a combination of these types of custom data.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "PhotoOverlay"
    _kml_fields = Overlay._kml_fields + (
        _FieldDef("shape"),
        _FieldDef("rotation"),
    )
    _kml_children = Overlay._kml_children + (
        _ChildDef("point"),
        _ChildDef("view_volume"),
        _ChildDef("image_pyramid"),
    )

    def __init__(
        self,
        # PhotoOverlay parameters
        point: Point,
        shape: OverlayShapeEnum | None = None,
        rotation: float | None = None,
        view_volume: ViewVolume | None = None,
        image_pyramid: ImagePyramid | None = None,
        # Overlay parameters
        icon: str | Icon | None = None,
        color: GeoColor | int | None = None,
        draw_order: int | None = None,
        # Feature parameters
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
    ):
        """PhotoOverlay instance constructor."""
        Overlay.__init__(
            self,
            icon=icon,
            color=color,
            draw_order=draw_order,
            name=name,
            visibility=visibility,
            is_open=is_open,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
        )
        self.point = point
        self.shape = shape
        self.rotation = rotation
        self.view_volume = view_volume
        self.image_pyramid = image_pyramid



"""Track module."""

from datetime import datetime
from typing import Any, Iterator, Iterable

from dateutil.parser import parse as dtparse
from lxml import etree  # type: ignore

from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Model import Model
from pyLiveKML.objects.Object import (
    _ChildDef,
    _FieldDef,
    _Angle180,
    _Angle360,
    _Angle90,
    _AnglePos180,
)
from pyLiveKML.types import AltitudeModeEnum
from pyLiveKML.utils import with_ns


class TrackCoord:
    """Track element coordinates class, specific to the `TrackElement` class.

    Used to collect the LLA (longitude/latitude/altitude) for a `TrackElement`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-track

    Parameters
    ----------
    lon: float, default = 0
    lat: float, default = 0
    alt: float, default = 0

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        lon: float = 0,
        lat: float = 0,
        alt: float = 0,
    ):
        """TrackCoord instance constructor."""
        super().__init__()
        self.lon = _Angle180.parse(lon)
        self.lat = _Angle90.parse(lat)
        self.alt = alt

    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.lon} {self.lat} {self.alt}"


class TrackAngles:
    """Track elements angles class, specific to the `TrackElement` class.

    Used to collect the rotation vector (heading/tilt/roll) for a `TrackElement`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-track

    Parameters
    ----------
    heading: float, default = 0
    tilt: float, default = 0
    roll: float, default = 0

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        heading: float = 0,
        tilt: float = 0,
        roll: float = 0,
    ) -> None:
        """TrackAngles instance constructor."""
        super().__init__()
        self.heading = _Angle360.parse(heading)
        self.tilt = _AnglePos180.parse(tilt)
        self.roll = _Angle180.parse(roll)

    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.heading} {self.tilt} {self.roll}"


class TrackElement:
    """Track element class, specific to the `Track` class.

    Used to collect the various components of each track point that are then distributed
    across the `<gx:Track>` tag when it is published.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-track

    Parameters
    ----------
    when : datetime | str | None, default = None
        The timestamp of the `TrackElement`. If a `str` is provided, it will be parsed
        with dateutils. Highly recommended to use an ISO format timestring.
    coords : TrackCoord | tuple[float, float, float] | tuple[float, float] | None, default = None
        The LLA coordinates of the track point, either a `TrackCoord` instance, or a
        tuple containing longitude, latitude and (optional) altitude, in that order.
    angles : TrackAngles | tuple[float, float, float] | tuple[float, float] | tuple[float] | float | None, default = None
        The rotational vector of the track point, either a `TrackAngles` instance, or a
        tuple containing heading, tilt and roll, in that order.
    extended_data: dict[str, dict[str, Any]] | None, default = None
        Extended data information. The key/s of the primary dict need to be # references
        for a `Schema` `id` in a `Document` hosting the `Track`. The sub-keys need to be
        fieldnames in their parent `Schema`. Technically, the

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self,
        when: datetime | str | None = None,
        coords: (
            TrackCoord | tuple[float, float, float] | tuple[float, float] | None
        ) = None,
        angles: (
            TrackAngles
            | tuple[float, float, float]
            | tuple[float, float]
            | tuple[float]
            | float
            | None
        ) = None,
        extended_data: dict[str, dict[str, Any]] | None = None,
    ) -> None:
        """TrackElement instance constructor."""
        super().__init__()
        self.when: datetime | None
        if isinstance(when, str):
            self.when = dtparse(when)
        else:
            self.when = when
        self.coords: TrackCoord | None = None
        if coords is None or isinstance(coords, TrackCoord):
            self.coords = coords
        else:
            self.coords = TrackCoord(*coords)
        self.angles: TrackAngles | None = None
        if angles is None or isinstance(angles, TrackAngles):
            self.angles = angles
        elif isinstance(angles, tuple):
            self.angles = TrackAngles(*angles)
        else:
            self.angles = TrackAngles(angles)
        self.extended_data = extended_data


class Track(Geometry):
    """A KML `<gx:Track>` tag constructor.

    A `Track` describes how an object moves through the world over a given time period.
    This feature allows you to create one visible object in Google Earth (either a
    `Point` icon or a `Model`) that encodes multiple positions for the same object for
    multiple times. In Google Earth, the time slider allows the user to move the view
    through time, which animates the position of the object.

    A `MultiTrack` is used to collect multiple tracks into one conceptual unit with one
    associated icon (or `Model`) that moves along the track. This feature is useful if
    you have multiple tracks for the same real-world object. The `interpolate` attribute
    of a `MultiTrack` specifies whether to interpolate between the tracks in a
    `MultiTrack`. If this value is `False`, then the point or `Model` stops at the end of
    one track and jumps to the start of the next one. (For example, if you want a single
    `Placemark` to represent your travels on two days, and your GPS unit was turned off
    for four hours during this period, you would want to show a discontinuity between the
    points where the unit was turned off and then on again.) If the value for
    `interpolate` is `True`, the values between the end of the first track and the
    beginning of the next track are interpolated so that the track appears as one
    continuous path.

    Notes
    -----
    * Although Google's documentation states that "sparse" data, i.e. extended data with
    missing values, is handled smoothly, I have been unable to replicate this in
    practice. I have found that the data trace for any schema field with a "sparse"
    dataset does not appear in the elevation profile.
    * As far as I have been able to establish, extended data **must** be numeric (int or
    float), or it does not appear in the elevation profile.
    * As far as I have been able to establish, extended data does not appear **at all**
    in the elevation profile for `MultiTrack` instances.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxtrack

    Parameters
    ----------
    altitude_mode : AltitudeModeEnum | None, default = None
        Specifies how altitude components in the `coordinates` attribute of each element
        are interpreted.
    elements : TrackElement | Iterable[TrackElement] | None, default = None
        The elements (points) along the `Track`.
    model : Model | None, default = None
        If specified, the `Model` replaces the `Point` icon used to indicate the current
        position on the track. When a `Model` is specified within a `Track`, it's
        attributes behave as follows:
        * The `location` attribute is ignored.
        * The `altitude_mode` attribute is ignored.
        * The `orientation` attribute is combined with the orientation of the track as
        follows. First, the `orientation` rotation is applied, which brings the `Model`
        from its local (x, y, z) coordinate system to a right-side-up, north-facing
        orientation. Next, a rotation is applied that corresponds to the interpolation
        of the `angles` values that affect the heading, tilt, and roll of the model as
        it moves along the track. If no angles are specified, the heading and tilt are
        inferred from the movement of the model.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:Track"
    _kml_fields = Geometry._kml_fields + (
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )
    _kml_children = Geometry._kml_children + (_ChildDef("model"),)

    def __init__(
        self,
        altitude_mode: AltitudeModeEnum | None = None,
        elements: TrackElement | Iterable[TrackElement] | None = None,
        model: Model | None = None,
    ) -> None:
        """Track instance constructor."""
        Geometry.__init__(self)
        self.altitude_mode = altitude_mode
        self.model = model
        self._schemas = dict[str, set[str]]()
        self._elements = list[TrackElement]()
        self.elements = elements

    @property
    def elements(self) -> Iterator[TrackElement]:
        """Retrieve a generator over the set of enclosed `TrackElements`."""
        yield from self._elements

    @elements.setter
    def elements(self, value: TrackElement | Iterable[TrackElement] | None) -> None:
        self._elements.clear()
        self._schemas.clear()
        if value is None:
            return
        if isinstance(value, TrackElement):
            self._elements.append(value)
        else:
            self._elements.extend(value)
            for xd in (
                e.extended_data for e in self._elements if e.extended_data is not None
            ):
                for schema, fields in xd.items():
                    if schema not in self._schemas:
                        self._schemas[schema] = set[str]()
                    self._schemas[schema].update(fields.keys())

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element."""
        super().build_kml(root, with_children)
        value: str
        for e in self.elements:
            value = "" if e.when is None else e.when.isoformat()
            etree.SubElement(root, with_ns("when")).text = value
        for e in self.elements:
            value = "" if e.coords is None else str(e.coords)
            etree.SubElement(root, with_ns("gx:coord")).text = value
        for e in self.elements:
            value = "" if e.angles is None else str(e.angles)
            etree.SubElement(root, with_ns("gx:angles")).text = value
        if self._schemas:
            e_xd = etree.SubElement(root, "ExtendedData")
            for sch, fields in self._schemas.items():
                e_sch = etree.SubElement(e_xd, "SchemaData", attrib={"schemaUrl": sch})
                for f in sorted(fields):
                    e_field = etree.SubElement(
                        e_sch,
                        with_ns("gx:SimpleArrayData"),
                        attrib={with_ns("kml:name"): f},
                    )
                    for e in self.elements:
                        value = ""
                        if (
                            e.extended_data
                            and sch in e.extended_data
                            and f in e.extended_data[sch]
                            and e.extended_data[sch][f] is not None
                        ):
                            value = str(e.extended_data[sch][f])
                        etree.SubElement(e_field, with_ns("gx:value")).text = value



"""Vec2 module."""

from abc import ABC

from lxml import etree  # type: ignore

from pyLiveKML.types.types import UnitsEnum
from pyLiveKML.objects.Object import _BaseObject


class Vec2(_BaseObject, ABC):
    """Abstract base for Vec2 subclasses.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    Parameters
    ----------
    x : float, default = 0.5
    y : float, default = 0.5
    x_units : UnitsEnum, default = UnitsEnum.FRACTION
    y_units : UnitsEnum, default = UnitsEnum.FRACTION

    """

    def __init__(
        self,
        x: float = 0.5,
        y: float = 0.5,
        x_units: UnitsEnum = UnitsEnum.FRACTION,
        y_units: UnitsEnum = UnitsEnum.FRACTION,
    ):
        """Vec2 instance constructor."""
        super().__init__()
        self.x = x
        self.y = y
        self.x_units = x_units
        self.y_units = y_units

    def construct_kml(
        self, with_children: bool = True, with_dependents: bool = True
    ) -> etree.Element:
        """Construct this instances' KML representation."""
        attribs = {
            "x": str(self.x),
            "y": str(self.y),
            "xunits": str(self.x_units.value),
            "yunits": str(self.y_units.value),
        }
        return etree.Element(self.kml_tag, attrib=attribs)

    def __eq__(self, other: object) -> bool:
        """Check equality."""
        return (
            isinstance(other, type(self))
            and self.x == other.x
            and self.y == other.y
            and self.x_units == other.x_units
            and self.y_units == other.y_units
        )


class HotSpot(Vec2):
    """Icon hotspot descriptor.

    Specifies the position within an `<Icon>` that is "anchored" to a `<Point>`.

    The x and y values can be specified in three different ways: as pixels, as fractions
    of the icon, or as inset pixels, which is an offset in pixels from the upper right
    corner of the icon. The x and y positions can be specified in different ways; for
    example, x can be in pixels and y can be a fraction. The origin of the coordinate
    system is in the lower left corner of the icon.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-iconstyle

    Parameters
    ----------
    x : float, default = 0.5
        Either the number of pixels, a fractional component of the icon, or a pixel inset
        indicating the x component of a point on the icon.
    y : float, default = 0.5
        Either the number of pixels, a fractional component of the icon, or a pixel inset
        indicating the y component of a point on the icon.
    x_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the x value is specified.
    y_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the y value is specified.

    """

    _kml_tag = "hotSpot"


class OverlayXY(Vec2):
    """Overlay image offset descriptor.

    Specifies a point on (or outside of) the overlay image that is mapped to the screen
    coordinate. It requires x and y values, and the units for those values.

    The x and y values can be specified in three different ways: as pixels, as fractions
    of the image, or as inset pixels, which is an offset in pixels from the upper right
    corner of the image. The x and y positions can be specified in different ways; for
    example, x can be in pixels and y can be a fraction. The origin of the coordinate
    system is in the lower left corner of the image.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-screenoverlay

    Parameters
    ----------
    x : float, default = 0.5
        Either the number of pixels, a fractional component of the image, or a pixel inset
        indicating the x component of a point on the image.
    y : float, default = 0.5
        Either the number of pixels, a fractional component of the image, or a pixel inset
        indicating the y component of a point on the image.
    x_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the x value is specified.
    y_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the y value is specified.

    """

    _kml_tag = "overlayXY"


class ScreenXY(Vec2):
    """Overlay screen offset descriptor.

    Specifies a point relative to the screen origin that the overlay image is mapped to.

    The x and y values can be specified in three different ways: as pixels, as fractions
    of the screen, or as inset pixels, which is an offset in pixels from the upper right
    corner of the screen. The x and y positions can be specified in different ways; for
    example, x can be in pixels and y can be a fraction. The origin of the coordinate
    system is in the lower left corner of the screen.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-screenoverlay

    Parameters
    ----------
    x : float, default = 0.5
        Either the number of pixels, a fractional component of the screen, or a pixel inset
        indicating the x component of a point on the screen.
    y : float, default = 0.5
        Either the number of pixels, a fractional component of the screen, or a pixel inset
        indicating the y component of a point on the screen.
    x_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the x value is specified.
    y_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the y value is specified.

    """

    _kml_tag = "screenXY"


class RotationXY(Vec2):
    """Overlay screen rotation descriptor.

    Point relative to the screen about which the screen overlay is rotated.

    The x and y values can be specified in three different ways: as pixels, as fractions
    of the screen, or as inset pixels, which is an offset in pixels from the upper right
    corner of the screen. The x and y positions can be specified in different ways; for
    example, x can be in pixels and y can be a fraction. The origin of the coordinate
    system is in the lower left corner of the screen.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-screenoverlay

    Parameters
    ----------
    x : float, default = 0.5
        Either the number of pixels, a fractional component of the screen, or a pixel inset
        indicating the x component of a point on the screen.
    y : float, default = 0.5
        Either the number of pixels, a fractional component of the screen, or a pixel inset
        indicating the y component of a point on the screen.
    x_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the x value is specified.
    y_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the y value is specified.

    """

    _kml_tag = "rotationXY"


class Size(Vec2):
    """Overlay screen size descriptor.

    Specifies the size of the image for the screen overlay. Separately for x and y:

    * A value of -1 indicates to use the native dimension
    * A value of 0 indicates to maintain the aspect ratio
    * A value of n sets the value of the dimension

    The x and y values can be specified in three different ways: as pixels, as fractions
    of the image, or as inset pixels, which is an offset in pixels from the upper right
    corner of the image. The x and y positions can be specified in different ways; for
    example, x can be in pixels and y can be a fraction. The origin of the coordinate
    system is in the lower left corner of the image.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-screenoverlay

    Parameters
    ----------
    x : float, default = 0.5
        Either the number of pixels, a fractional component of the screen, or a pixel inset
        indicating the x component of a point on the screen.
    y : float, default = 0.5
        Either the number of pixels, a fractional component of the screen, or a pixel inset
        indicating the y component of a point on the screen.
    x_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the x value is specified.
    y_units : UnitsEnum, default = UnitsEnum.FRACTION
        Units in which the y value is specified.

    """

    _kml_tag = "size"



"""Schema module."""

from typing import Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import (
    _BaseObject,
    _DependentDef,
    _FieldDef,
    Object,
)


class SimpleField(_BaseObject):
    """SimpleField class.

    `Schema` instances contain a collection of `SimpleField` instances.
    """

    _kml_tag = "SimpleField"
    _kml_fields = _BaseObject._kml_fields + (_FieldDef("display_name", "displayName"),)

    def __init__(
        self,
        type: str,
        name: str,
        display_name: str | None = None,
    ) -> None:
        """SimpleField instance constructor."""
        super().__init__()
        self.type = type
        self.name = name
        self.display_name = display_name

    def construct_kml(
        self, with_children: bool = True, with_dependents: bool = True
    ) -> etree.Element:
        """Construct this instances' KML representation."""
        root = etree.Element(
            self.kml_tag, attrib={"type": self.type, "name": self.name}
        )
        self.build_kml(root, with_children, with_dependents)
        return root


class Schema(Object):
    """A KML `<Schema>` tag constructor.

    Specifies a custom KML schema that is used to add custom data to KML `Feature`s.
    `Schema` is always a child of `Document`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#schema

    Parameters
    ----------
    name : str
        The name of the `Schema`.
    schema_fields : SimpleField | Iterable[SimpleField] | None, default = None
        The fields of the `Schema`.

    Attributes
    ----------
    Same as parameters

    """

    _kml_tag = "Schema"
    _kml_dependents = Object._kml_dependents + (_DependentDef("schema_fields"),)

    def __init__(
        self,
        name: str,
        schema_fields: SimpleField | Iterable[SimpleField] | None = None,
    ) -> None:
        """Construct Schema instances."""
        Object.__init__(self)
        self.name = name
        self._schema_fields = list[SimpleField]()
        self.schema_fields = schema_fields

    @property
    def schema_fields(self) -> Iterator[SimpleField]:
        """Generator over schema_fields."""
        yield from self._schema_fields

    @schema_fields.setter
    def schema_fields(self, value: SimpleField | Iterable[SimpleField] | None) -> None:
        self._schema_fields.clear()
        if value is not None:
            if isinstance(value, SimpleField):
                self._schema_fields.append(value)
            else:
                self._schema_fields.extend(value)

    def construct_kml(
        self, with_children: bool = True, with_dependents: bool = True
    ) -> etree.Element:
        """Construct this instances' KML representation.

        Adds the `name` attribute to the returned tag.
        """
        root = super().construct_kml(with_children, with_dependents)
        root.set("name", self.name)
        return root



"""GeoCoordinates module."""

from typing import Any

from pyLiveKML.objects.Object import _Angle90, _Angle180


class GeoCoordinates:
    """The GeoCoordinates type describes a single instance of a Lon-Lat-Alt (LLA) position.

    Notes
    -----
    The GeoCoordinates type is *not* explicitly referenced by the KML specification;
    rather, it is a construct of convenience for the pyLiveKML package.

    Parameters
    ----------
    lon : float
        The longitude, in decimal degrees
    lat : float
        The latitude, in decimal degrees
    alt : float | None, default = None
        The altitude, in metres

    Attributes
    ----------
    lon : float
        The longitude, in decimal degrees
    lat : float
        The latitude, in decimal degrees
    alt : float | None, default = None
        The altitude, in metres

    """

    def __init__(
        self,
        lon: float = 0,
        lat: float = 0,
        alt: float | None = None,
    ):
        """GeoCoordinates instance constructor."""
        self.lon = lon
        self.lat = lat
        self.alt = alt

    @property
    def values(self) -> tuple[float, float, float | None]:
        """Get or set the LLA as a tuple.

        Parameters
        ----------
        values : tuple[float, float, float|None] | tuple[float, float]
            The LLA as a 2- or 3-tuple, in lon-lat[-alt] order.

        Returns
        -------
        tuple[float, float, float|None]
            The LLA as a tuple

        """
        return (self.lon, self.lat, self.alt)

    @values.setter
    def values(
        self, value: tuple[float, float, float | None] | tuple[float, float]
    ) -> None:
        if len(value) == 3:
            self.lon, self.lat, self.alt = value
        else:
            self.lon, self.lat, self.alt = (*value[:2], None)

    def __setattr__(self, name: str, value: Any) -> None:
        """GeoCoordinates __setattr__ implementation."""
        if name == "lon":
            value = _Angle180.parse(value)
        elif name == "lat":
            value = _Angle90.parse(value)
        super().__setattr__(name, value)

    def __str__(self) -> str:
        """Return a string representation."""
        if self.alt is None:
            return f"{self.lon:0.6f},{self.lat:0.6f}"
        else:
            return f"{self.lon:0.6f},{self.lat:0.6f},{self.alt:0.1f}"



"""Camera module."""

from typing import Iterable, cast

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView, ViewerOption
from pyLiveKML.objects.Object import (
    _FieldDef,
    _Angle90,
    _Angle180,
    _AnglePos180,
    _Angle360,
)
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types import AltitudeModeEnum, GeoCoordinates, ViewerOptionEnum


class Camera(AbstractView):
    """A KML `<Camera>` tag constructor.

    Defines the virtual camera that views the scene. This element defines the position of
    the camera relative to the Earth's surface as well as the viewing direction of the
    camera. The camera position is defined by `<longitude>`, `<latitude>`, `<altitude>`,
    and either `<altitudeMode>` or `<gx:altitudeMode>`. The viewing direction of the
    camera is defined by `<heading>`, `<tilt>`, and `<roll>`. `<Camera>` can be a child
    element of any `Feature` or of `<NetworkLinkControl>`. A parent element cannot
    contain both a `<Camera>` and a `<LookAt>` at the same time.

    `Camera` provides full six-degrees-of-freedom control over the view, so you can
    position the Camera in space and then rotate it around the X, Y, and Z axes. Most
    importantly, you can tilt the camera view so that you're looking above the horizon
    into the sky.

    `Camera` can also contain a :class:`pyLiveKML.objects.TimePrimitive`
    (`GxTimeSpan` or `GxTimeStamp`, noting the "Gx" variant). Time values in `Camera`
    affect historical imagery, sunlight, and the display of time-stamped features.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#camera
    * https://developers.google.com/kml/documentation/time#abstractviews

    Parameters
    ----------
    lla : GeoCoordinates | tuple[float, float, float|None] | tuple[float, float] | None, default = None
        The longitude, latitude and altitude (in that order, in decimal degrees) of the
        `Camera` position. If `None`, `lon` and `lat` must be supplied or an exception
        will be thrown.
    heading : float, default = 0
        The heading (in decimal degrees) of the `Camera` facing.
    tilt : float, default = 0
        The tilt (in decimal degrees) of the `Camera` facing.
    roll : float, default = 0
        The roll (in decimal degrees) of the `Camera` facing.
    altitude_mode : AltitudeModeEnum | None, default = None
    lon : float | None, default = None
        The longitude (in decimal degrees) of the the `Camera` position. If `None`, `lla`
        must be supplied.
    lat : float | None, default = None
        The latitude (in decimal degrees) of the the `Camera` position. If `None`, `lla`
        must be supplied.
    alt : float | None, default = None
        The altitude (in metres, with respect to `altitude_mode`) of the the `Camera`
        position.
    viewer_options : ViewerOption | Iterable[ViewerOption] | None, default = None
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None, default = None
        `TimeStamp` or `TimeSpan` assigned to the object.

    Attributes
    ----------
    viewer_options : list[ViewerOption]
    time_primitive : TimePrimitive | None
    lon : float
        The longitude of the camera position, in decimal degrees.
    lat : float
        The latitude of the camera position, in decimal degrees.
    alt : float
        The altitude of the camera position, in metres, with respect to `altitude_mode`.
    heading : float
        The heading of the camera facing, in decimal degrees.
    tilt : float
        The tilt of the camera facing, in decimal degrees.
    roll : float
        The roll of the camera facing, in decimal degrees.
    altitude_mode : AltitudeModeEnum | None

    """

    _kml_tag = "Camera"
    _kml_fields = AbstractView._kml_fields + (
        _FieldDef("lon", "longitude", parser=_Angle180),
        _FieldDef("lat", "latitude", parser=_Angle90),
        _FieldDef("alt", "altitude"),
        _FieldDef("heading", parser=_Angle360),
        _FieldDef("tilt", parser=_AnglePos180),
        _FieldDef("roll", parser=_Angle180),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )

    def __init__(
        self,
        lla: (
            GeoCoordinates
            | tuple[float, float, float | None]
            | tuple[float, float]
            | None
        ) = None,
        heading: float = 0,
        tilt: float = 0,
        roll: float = 0,
        altitude_mode: AltitudeModeEnum | None = None,
        lon: float | None = None,
        lat: float | None = None,
        alt: float | None = None,
        viewer_options: Iterable[ViewerOption] | ViewerOption | None = None,
        time_primitive: TimePrimitive | None = None,
    ):
        """Camera instance constructor."""
        AbstractView.__init__(self, viewer_options, time_primitive)
        self.lon: float
        self.lat: float
        self.alt: float
        if lla is None and (lon is None or lat is None):
            raise ValueError(
                "You must supply either `lla` or `lon` and `lat` (and optionally `alt`)."
            )
        if lla is None:
            self.lon = cast(float, lon)
            self.lat = cast(float, lat)
            self.alt = cast(float, alt)
        else:
            self.lla = lla
        self.heading, self.tilt, self.roll = heading, tilt, roll
        self.altitude_mode = altitude_mode

    @property
    def lla(self) -> tuple[float, float, float]:
        """Get or set the longitude, latitude and altitude as a tuple.

        Parameters
        ----------
        value : GeoCoordinates | tuple[float, float, float|None] | tuple[float, float])
            The longitude, latitude and altitude as a tuple with optional altitude, or as
            `GeoCoordinates`.

        Returns
        -------
        tuple[float, float, float]
            The longitude, latitude and altitude as a tuple of floats.

        """
        return (self.lon, self.lat, self.alt)

    @lla.setter
    def lla(
        self,
        value: GeoCoordinates | tuple[float, float, float | None] | tuple[float, float],
    ) -> None:
        if isinstance(value, GeoCoordinates):
            v = value.values
            self.lon, self.lat = v[:2]
            self.alt = 0 if v[2] is None else v[2]
        elif len(value) >= 3:
            v = value[:3]
            self.lon, self.lat = v[:2]
            self.alt = 0 if v[2] is None else v[2]
        else:
            self.lon, self.lat = value
            self.alt = 0



"""ColorStyle module."""

from abc import ABC

from pyLiveKML.objects.Object import _FieldDef, _ColorParse
from pyLiveKML.objects.SubStyle import SubStyle
from pyLiveKML.types import ColorModeEnum, GeoColor


class ColorStyle(SubStyle, ABC):
    """A KML `<ColorStyle>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file. It provides
    elements for specifying the color and color mode of extended style types.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#colorstyle

    Parameters
    ----------
    color: GeoColor | int | None, default = None
    color_mode: ColorModeEnum | None, default = None

    Attributes
    ----------
    Same as parameters.

    """

    _kml_fields: tuple[_FieldDef, ...] = SubStyle._kml_fields + (
        _FieldDef("color", parser=_ColorParse),
        _FieldDef("color_mode", "colorMode"),
    )

    def __init__(
        self,
        color: GeoColor | int | None = None,
        color_mode: ColorModeEnum | None = None,
    ):
        """ColorStyle instance constructor."""
        SubStyle.__init__(self)
        ABC.__init__(self)
        self.color_mode = color_mode
        self.color = GeoColor(color) if isinstance(color, int) else color



"""FlyTo module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.Object import _ChildDef, _FieldDef
from pyLiveKML.objects.TourPrimitive import TourPrimitive
from pyLiveKML.types import FlyToModeEnum


class FlyTo(TourPrimitive):
    """A KML `<gx:FlyTo>` tag constructor.

    Specifies a point in space to which the browser will fly during a tour. It must
    contain one `<AbstractView>`, and should contain `<gx:duration>` and `<gx:flyToMode>`
    elements.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxflyto.

    Parameters
    ----------
    duration : float
        The time (in seconds) it takes to fly to the defined point from the current point.
    fly_to_mode : FlyToModeEnum
        The method of flight to the defined point from the current point.
    abstract_view : AbstractView
        A concrete `AbstractView` subclass that describes the view while transitioning to
        the defined point from the current point.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:FlyTo"
    _kml_fields = TourPrimitive._kml_fields + (
        _FieldDef("duration", "gx:duration"),
        _FieldDef("fly_to_mode", "gx:flyToMode"),
    )
    _kml_children = TourPrimitive._kml_children + (_ChildDef("abstract_view"),)

    def __init__(
        self,
        duration: float,
        fly_to_mode: FlyToModeEnum,
        abstract_view: AbstractView,
    ) -> None:
        """Track instance constructor."""
        TourPrimitive.__init__(self)
        self.duration = duration
        self.fly_to_mode = fly_to_mode
        self.abstract_view = abstract_view



"""Polygon module."""

from typing import Iterable, Iterator, cast

from lxml import etree  # type: ignore

from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.LinearRing import LinearRing
from pyLiveKML.objects.Object import _FieldDef, _DependentDef, Object, ObjectState
from pyLiveKML.types import AltitudeModeEnum, GeoCoordinates


PolyBoundaryType = (
    LinearRing
    | Iterable[GeoCoordinates]
    | Iterable[tuple[float, float, float]]
    | Iterable[tuple[float, float]]
)


def _is_boundary_iterable(val: PolyBoundaryType | Iterable[PolyBoundaryType]) -> bool:
    """Test whether the argument is a PolyBoundaryType, or an iterable of PolyBoundaryType."""
    # if it's a LinearRing, then it's singular
    if isinstance(val, LinearRing):
        return False
    # if it's not a LinearRing, then it _must_ be iterable - further checks
    first_it = iter(val)
    first_obj = next(first_it, None)
    # if it's an iterable of LinearRings, then it's multiple
    if isinstance(first_obj, LinearRing):
        return True
    # if it's an iterable of GeoCoordinates, then it's singular
    if isinstance(first_obj, GeoCoordinates):
        return False
    # if it's an iterable, then it could be either - further checks
    if isinstance(first_obj, Iterable):
        second_it = iter(first_obj)
        second_obj = next(second_it, None)
        # if it's an iterable of floats, then its singular
        if isinstance(second_obj, float):
            return False
        # if it's an iterable of GeoCoordinate or tuples, then it's multiple
        if isinstance(second_obj, (GeoCoordinates, tuple)):
            return True
    # if we arrive here, it hasn't been recognized; raise a ValueError
    raise ValueError(
        "Unable to determine whether boundary argument is singular or multiple."
    )


class _OuterBoundary(Object):
    """Private wrapper class for polygon outer boundaries."""

    _kml_tag = "outerBoundaryIs"
    _kml_dependents = Object._kml_dependents + (_DependentDef("boundary"),)
    _suppress_id = True

    def __init__(self, boundary: PolyBoundaryType) -> None:
        super().__init__()
        if isinstance(boundary, LinearRing):
            self.boundary = boundary
        else:
            self.boundary = LinearRing(boundary)


class _InnerBoundary(Object):
    """Private wrapper class for polygon inner boundaries."""

    _kml_tag = "innerBoundaryIs"
    _kml_dependents = Object._kml_dependents + (_DependentDef("boundary"),)
    _suppress_id = True

    def __init__(self, boundary: PolyBoundaryType) -> None:
        super().__init__()
        if isinstance(boundary, LinearRing):
            self.boundary = boundary
        else:
            self.boundary = LinearRing(boundary)


class Polygon(Geometry):
    """A KML `<Polygon>` tag constructor.

    A `Polygon` is defined by an outer boundary and 0 or more inner boundaries. The
    boundaries, in turn, are defined by `LinearRing`s. When a `Polygon` is extruded, its
    boundaries are connected to the ground to form additional polygons, which gives the
    appearance of a building or a box. Extruded `Polygon`s use `PolyStyle` for their
    `color`, `color mode`, and `fill`.

    The `coordinates` for `Polygon`s (or, more precisely, the `LinearRing`s that make up
    `Polygon`s) **must** be specified in counterclockwise order. `Polygon`s follow the
    "right-hand rule," which states that if you place the fingers of your right hand in
    the direction in which the coordinates are specified, your thumb points in the
    general direction of the geometric normal for the `Polygon`. (In 3D graphics, the
    geometric normal is used for lighting and points away from the front face of the
    `Polygon`.) Since Google Earth fills only the front face of `Polygon`s, you will
    achieve the desired effect only when the coordinates are specified in the proper
    order. Otherwise, the `Polygon` will be gray.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#polygon.

    Typing
    ------
    PolyBoundaryType : LinearRing | Iterable[GeoCoordinates] | Iterable[tuple[float, float, float]] | Iterable[tuple[float, float]]

    Parameters
    ----------
    outer_boundary : PolyBoundaryType
    inner_boundaries : PolyBoundaryType | Iterable[PolyBoundaryType] | None = None
    altitude_mode : AltitudeModeEnum | None = None
        Specifies how `altitude` attributes in the `coordinates` attribute are
        interpreted.
    extrude : bool | None = None
        Specifies whether to connect the Polygon to the ground. Only the vertices are
        extruded, not the geometry itself; for example, a rectangle turns into a box with
        five faces. The vertices of the `Polygon` are extruded toward the center of the
        Earth's geoid.
    tessellate : bool | None = None
        This field is not used by Polygon.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Polygon"
    _kml_fields = Geometry._kml_fields + (
        _FieldDef("altitude_mode", "gx:altitudeMode"),
        _FieldDef("extrude"),
        _FieldDef("tessellate"),
    )
    _kml_dependents = Geometry._kml_dependents + (
        _DependentDef("_outer_boundary"),
        _DependentDef("_inner_boundaries"),
    )

    def __init__(
        self,
        outer_boundary: PolyBoundaryType,
        inner_boundaries: PolyBoundaryType | Iterable[PolyBoundaryType] | None = None,
        altitude_mode: AltitudeModeEnum | None = None,
        extrude: bool | None = None,
        tessellate: bool | None = None,
    ):
        """Polygon instance constructor."""
        Geometry.__init__(self)
        self.outer_boundary = outer_boundary
        self._inner_boundaries = list[_InnerBoundary]()
        self.inner_boundaries = inner_boundaries
        self.extrude = extrude
        self.tessellate = tessellate
        self.altitude_mode = altitude_mode

    @property
    def outer_boundary(self) -> LinearRing:
        """Outer boundary."""
        return self._outer_boundary.boundary

    @outer_boundary.setter
    def outer_boundary(self, value: PolyBoundaryType) -> None:
        self._outer_boundary = _OuterBoundary(value)
        if self.active:
            self._outer_boundary.boundary._state = ObjectState.CHANGING

    @property
    def inner_boundaries(self) -> Iterator[LinearRing]:
        """Generator across inner boundaries."""
        for b in self._inner_boundaries:
            yield b.boundary

    @inner_boundaries.setter
    def inner_boundaries(
        self, value: PolyBoundaryType | Iterable[PolyBoundaryType] | None
    ) -> None:
        self._inner_boundaries.clear()
        if value:
            if not _is_boundary_iterable(value):
                self._inner_boundaries.append(
                    _InnerBoundary(cast(PolyBoundaryType, value))
                )
            else:
                self._inner_boundaries.extend(
                    map(_InnerBoundary, cast(Iterable[PolyBoundaryType], value))
                )
        if self.active:
            for b in self._inner_boundaries:
                b.boundary._state = ObjectState.CHANGING



"""StyleMap module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _BaseObject, _ChildDef, _DependentDef, _FieldDef
from pyLiveKML.objects.Style import Style
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.types.types import StyleStateEnum


class _StyleMap_Pair(_BaseObject):
    """Hidden class for use by `StyleMap`."""

    _kml_tag = "Pair"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("key"),
        _FieldDef("style_url", "styleUrl"),
    )
    _kml_dependents = _BaseObject._kml_dependents + (_DependentDef("style"),)

    def __init__(self, key: StyleStateEnum, style_ref: str | Style):
        """_StyleMap_Pair instance constructor."""
        super().__init__()
        self.key = key
        self.style: Style | None
        self.style_url: str | None
        if isinstance(style_ref, Style):
            self.style = style_ref
            self.style_url = None
        else:
            self.style = None
            self.style_url = style_ref

    @property
    def value(self) -> str | Style:
        return self.style_url if self.style_url else self.style if self.style else ""

    @value.setter
    def value(self, value: str | Style) -> None:
        if isinstance(value, str):
            self.style = None
            self.style_url = value
        else:
            self.style = value
            self.style_url = None


class StyleMap(StyleSelector):
    """A KML `<StyleMap>` tag constructor.

    A `StyleMap` maps between two different `Style`s. Typically a `StyleMap` is used to
    provide separate normal and highlighted styles for a `Placemark`, so that the
    highlighted version appears when the user mouses over the icon in Google Earth.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#stylemap

    Parameters
    ----------
    normal_style_ref : str | Style
        The "normal" style, displayed when the mouse **is not** hovering over the
        `Feature`.
    highlight_style_ref : str | Style
        The "highlighted" style, displayed when the mouse **is** hovering over the
        `Feature`.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "StyleMap"
    _kml_children = StyleSelector._kml_children + (
        _ChildDef("_normal"),
        _ChildDef("_highlight"),
    )

    def __init__(
        self,
        normal_style_ref: str | Style,
        highlight_style_ref: str | Style,
    ):
        """StyleMap instance constructor."""
        super().__init__()
        self._normal = _StyleMap_Pair(StyleStateEnum.NORMAL, normal_style_ref)
        self._highlight = _StyleMap_Pair(StyleStateEnum.HIGHLIGHT, highlight_style_ref)

    @property
    def normal(self) -> str | Style:
        """Normal style reference or instance.

        Parameters
        ----------
        value : str | Style

        Returns
        -------
        str | Style

        """
        return self._normal.value

    @normal.setter
    def normal(self, value: str | Style) -> None:
        self._normal.value = value

    @property
    def highlight(self) -> str | Style:
        """Highlight style reference or instance.

        Parameters
        ----------
        value : str | Style

        Returns
        -------
        str | Style

        """
        return self._highlight.value

    @highlight.setter
    def highlight(self, value: str | Style) -> None:
        self._highlight.value = value



"""TourPrimitive module."""

from abc import ABC

from pyLiveKML.objects.Object import Object


class TourPrimitive(Object, ABC):
    """A KML `<TourPrimitive>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file.

    Objects inheriting from `TourPrimitive` provide instructions to KML browsers during
    tours, including points to fly to and the duration of those flights, pauses, updates
    to KML features, and sound files to play.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#tourprimitive

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    def __init__(self) -> None:
        """TourPrimitive instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)



"""pyLiveKML init module."""

from pyLiveKML.utils import (
    kml_root_tag,
    KML_DOCTYPE,
    KML_HEADERS,
    KML_UPDATE_CONTAINER_LIMIT_DEFAULT,
)
from pyLiveKML.types import (
    AltitudeModeEnum,
    ColorModeEnum,
    DisplayModeEnum,
    FlyToModeEnum,
    GeoColor,
    GeoCoordinates,
    HotSpot,
    ItemIconModeEnum,
    ListItemTypeEnum,
    OverlayShapeEnum,
    OverlayXY,
    PlayModeEnum,
    RefreshModeEnum,
    RotationXY,
    ScreenXY,
    Size,
    StyleStateEnum,
    UnitsEnum,
    ViewRefreshModeEnum,
    ViewerOptionEnum,
)
from pyLiveKML import errors
from pyLiveKML.objects.AbstractView import ViewerOption
from pyLiveKML.objects.AnimatedUpdate import AnimatedUpdate
from pyLiveKML.objects.BalloonStyle import BalloonStyle
from pyLiveKML.objects.Camera import Camera
from pyLiveKML.objects.Document import Document
from pyLiveKML.objects.ExtendedData import DataItem, ExtendedData, SchemaDataItem
from pyLiveKML.objects.FlyTo import FlyTo
from pyLiveKML.objects.Folder import Folder
from pyLiveKML.objects.GroundOverlay import GroundOverlay, LatLonBox, LatLonQuad
from pyLiveKML.objects.Icon import Icon
from pyLiveKML.objects.IconStyle import IconStyle
from pyLiveKML.objects.LabelStyle import LabelStyle
from pyLiveKML.objects.LineString import LineString
from pyLiveKML.objects.LineStyle import LineStyle
from pyLiveKML.objects.LinearRing import LinearRing
from pyLiveKML.objects.Link import Link
from pyLiveKML.objects.ListStyle import ItemIcon, ListStyle
from pyLiveKML.objects.LookAt import LookAt
from pyLiveKML.objects.Model import Alias, Model
from pyLiveKML.objects.MultiGeometry import MultiGeometry
from pyLiveKML.objects.MultiTrack import MultiTrack
from pyLiveKML.objects.NetworkLink import NetworkLink
from pyLiveKML.objects.NetworkLinkControl import NetworkLinkControl
from pyLiveKML.objects.PhotoOverlay import ImagePyramid, PhotoOverlay, ViewVolume
from pyLiveKML.objects.Placemark import Placemark
from pyLiveKML.objects.Point import Point
from pyLiveKML.objects.PolyStyle import PolyStyle
from pyLiveKML.objects.Polygon import Polygon
from pyLiveKML.objects.Region import LatLonAltBox, LevelOfDetail, Region
from pyLiveKML.objects.Schema import Schema, SimpleField
from pyLiveKML.objects.ScreenOverlay import ScreenOverlay
from pyLiveKML.objects.SoundCue import SoundCue
from pyLiveKML.objects.Style import Style
from pyLiveKML.objects.StyleMap import StyleMap
from pyLiveKML.objects.TimeSpan import TimeSpan
from pyLiveKML.objects.TimeStamp import TimeStamp
from pyLiveKML.objects.Tour import Tour, Playlist
from pyLiveKML.objects.TourControl import TourControl
from pyLiveKML.objects.Track import Track, TrackAngles, TrackCoord, TrackElement
from pyLiveKML.objects.Update import Update, UpdateSequent, UpdateType
from pyLiveKML.objects.Wait import Wait

# global imports wrapper
# allows importing any of the instantiable KML objects and helpers from the `pyLiveKML` module
__all__ = [
    "Alias",
    "AltitudeModeEnum",
    "AltitudeModeEnum",
    "AnimatedUpdate",
    "BalloonStyle",
    "Camera",
    "ColorModeEnum",
    "DataItem",
    "DisplayModeEnum",
    "Document",
    "ExtendedData",
    "FlyTo",
    "FlyToModeEnum",
    "Folder",
    "GeoColor",
    "GeoCoordinates",
    "GroundOverlay",
    "HotSpot",
    "Icon",
    "IconStyle",
    "ImagePyramid",
    "ItemIcon",
    "ItemIconModeEnum",
    "KML_DOCTYPE",
    "KML_HEADERS",
    "KML_UPDATE_CONTAINER_LIMIT_DEFAULT",
    "LabelStyle",
    "LatLonAltBox",
    "LatLonBox",
    "LatLonQuad",
    "LevelOfDetail",
    "LineString",
    "LineStyle",
    "LinearRing",
    "Link",
    "ListItemTypeEnum",
    "ListStyle",
    "LookAt",
    "Model",
    "MultiGeometry",
    "MultiTrack",
    "NetworkLink",
    "NetworkLinkControl",
    "OverlayShapeEnum",
    "OverlayXY",
    "PhotoOverlay",
    "Placemark",
    "PlayModeEnum",
    "Playlist",
    "Point",
    "PolyStyle",
    "Polygon",
    "RefreshModeEnum",
    "Region",
    "RotationXY",
    "Schema",
    "SchemaDataItem",
    "ScreenOverlay",
    "ScreenXY",
    "SimpleField",
    "Size",
    "SoundCue",
    "Style",
    "StyleMap",
    "StyleStateEnum",
    "TimeSpan",
    "TimeSpan",
    "TimeStamp",
    "Tour",
    "TourControl",
    "Track",
    "TrackAngles",
    "TrackCoord",
    "TrackElement",
    "UnitsEnum",
    "Update",
    "UpdateSequent",
    "UpdateType",
    "ViewRefreshModeEnum",
    "ViewerOption",
    "ViewerOptionEnum",
    "ViewVolume",
    "Wait",
    "errors",
    "kml_root_tag",
]



"""GeoColor module."""


class GeoColor:
    """The GeoColor type describes an ABGR-format color, as used throughout GEP.

    Notes
    -----
    * The GeoColor type is *not* explicitly referenced by the KML specification; rather,
    it is a construct of convenience for the pyLiveKML package.
    * When constructing a GeoColor instance, at least *one* of the parameters must be
    supplied.
    * Parameter values less than the lower bound (0) of the target component (r, g, b, a)
    will be set to the lower bound. Any positive value will be bit-masked with 0xff to
    constrain it to an 8-bit range (0-255).

    Parameters
    ----------
    value : int | tuple[int, int, int, int] | None, default = None
        The integral (32-bit) numeric value of the color, or a 4-tuple of integers in
        ABGR order.
    rgba : tuple[int, int, int, int] | None, default = None
        The color as a 4-tuple of integers, in RGBA order, i.e. reversed from `value`.

    Raises
    ------
    ValueError
        If all of the parameters are `None`.

    """

    def __init__(
        self,
        value: int | tuple[int, int, int, int] | None = None,
        rgba: tuple[int, int, int, int] | None = None,
    ):
        """GeoColor instance constructor."""
        self._r: int = 0
        self._g: int = 0
        self._b: int = 0
        self._a: int = 0
        if value is not None:
            if isinstance(value, int):
                self.value = value
            else:
                self.a, self.b, self.g, self.r = value
        elif rgba is not None:
            self.r, self.g, self.b, self.a = rgba
        else:
            raise ValueError("At least one of the parameters must not be `None`")

    @property
    def value(self) -> int:
        """Combined ABGR components (32 bits, in that order, MSB to LSB)."""
        return self._r | (self._g << 8) | (self._b << 16) | (self._a << 24)

    @value.setter
    def value(self, value: int) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else value
        value = 0xFFFFFFFF if value > 0xFFFFFFFF else value
        self._r = value & 0xFF
        self._g = (value >> 8) & 0xFF
        self._b = (value >> 16) & 0xFF
        self._a = (value >> 24) & 0xFF

    @property
    def bgr(self) -> int:
        """BGR portion (24 bits)."""
        return self._r + (self._g << 8) + (self._b << 16)

    @bgr.setter
    def bgr(self, value: int | None) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else 0xFFFFFF if value > 0xFFFFFF else value
        self._r = value & 0xFF
        self._g = (value >> 8) & 0xFF
        self._b = (value >> 16) & 0xFF

    @property
    def rgb(self) -> int:
        """RGB (24 bits, reversed from GEP format!) portion."""
        return self._b + (self._g << 8) + (self._r << 16)

    @rgb.setter
    def rgb(self, value: int | None) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else 0xFFFFFF if value > 0xFFFFFF else value
        self._b = value & 0xFF
        self._g = (value >> 8) & 0xFF
        self._r = (value >> 16) & 0xFF

    @property
    def r(self) -> int:
        """Red component (8 bits)."""
        return self._r

    @r.setter
    def r(self, value: int | None) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else 0xFF if value > 0xFF else value
        self._r = value

    @property
    def g(self) -> int:
        """Green component (8 bits)."""
        return self._g

    @g.setter
    def g(self, value: int | None) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else 0xFF if value > 0xFF else value
        self._g = value

    @property
    def b(self) -> int:
        """Blue component (8 bits)."""
        return self._b

    @b.setter
    def b(self, value: int | None) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else 0xFF if value > 0xFF else value
        self._b = value

    @property
    def a(self) -> int:
        """Alpha component (8 bits)."""
        return self._a

    @a.setter
    def a(self, value: int | None) -> None:
        value = 0 if value is None else value
        value = 0 if value < 0 else 0xFF if value > 0xFF else value
        self._a = value

    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.value:08x}"

    def __eq__(self, other: object) -> bool:
        """Check equality."""
        return isinstance(other, GeoColor) and self.value == other.value
